try:
    import wandb
except ImportError:
    wandb = None
import os
import os.path as osp
from argparse import ArgumentParser
from collections import defaultdict
from pprint import pprint
from typing import Any, Callable, Dict, List, Optional

import numpy as np
import pandas as pd
from omegaconf import DictConfig, OmegaConf

from rl_utils.common.core_utils import CacheHelper
from rl_utils.plotting.utils import MISSING_VALUE


def extract_query_key(k):
    if k.startswith("ALL_"):
        return k.split("ALL_")[1]
    return k


def batch_query(
    all_select_fields: List[List[str]],
    all_filter_fields: List[Dict[str, Any]],
    proj_cfg: Dict[str, Any],
    all_should_skip: Optional[List[bool]] = None,
    all_add_info: Optional[List[Dict[str, Any]]] = None,
    verbose=True,
    limit=None,
    use_cached=False,
    reduce_op: Optional[Callable[[List], float]] = None,
    error_ok: bool = False,
):
    """
    - all_should_skip: Whether to skip querying this value.
    """
    n_query = len(all_select_fields)
    if all_add_info is None:
        all_add_info = [None for _ in range(n_query)]
    if all_should_skip is None:
        all_should_skip = [False for _ in range(n_query)]

    data = []
    for select_fields, filter_fields, should_skip, add_info in zip(
        all_select_fields, all_filter_fields, all_should_skip, all_add_info
    ):
        r = []
        if not should_skip:
            r = query(
                select_fields,
                filter_fields,
                proj_cfg,
                verbose,
                limit,
                use_cached,
                reduce_op,
                error_ok=error_ok,
            )
        if len(r) == 0:
            r = [{k: MISSING_VALUE for k in select_fields}]
        for d in r:
            if add_info is None:
                data.append(d)
            else:
                data.append({**add_info, **d})
    return data


def query(
    select_fields: List[str],
    filter_fields: Dict[str, str],
    proj_cfg: Dict[str, Any],
    verbose=True,
    limit=None,
    use_cached=False,
    reduce_op: Optional[Callable[[List], float]] = None,
    error_ok: bool = False,
):
    """
    :param select_fields: The list of data to retrieve. If a field starts with
        "ALL_", then all the entries for this name from W&B are fetched. This gets
        the ENTIRE history. Other special keys include: "_runtime" (in
        seconds), "_timestamp".
    :param filter_fields: Key is the filter type (like group or tag) and value
        is the filter value (like the name of the group or tag to match)
    :param reduce_op: `np.mean` would take the average of the results.
    :param use_cached: Saves the results to disk so next time the same result is requested, it is loaded from disk rather than W&B.

    See README for more information.
    """

    wb_proj_name = proj_cfg["proj_name"]
    wb_entity = proj_cfg["wb_entity"]

    lookup = f"{select_fields}_{filter_fields}"
    cache = CacheHelper("wb_queries", lookup)

    if use_cached and cache.exists():
        return cache.load()
    if wandb is None:
        raise ValueError("Wandb is not installed")

    api = wandb.Api()

    query_dict = {}
    search_id = None

    for f, v in filter_fields.items():
        if f == "group":
            query_dict["group"] = v
        elif f == "tag":
            query_dict["tags"] = v
        elif f == "id":
            search_id = v
        else:
            query_dict["config." + f] = v

    def log(s):
        if verbose:
            print(s)

    if search_id is None:
        log("Querying with")
        log(query_dict)
        runs = api.runs(f"{wb_entity}/{wb_proj_name}", query_dict)
    else:
        log(f"Searching for ID {search_id}")
        runs = [api.run(f"{wb_entity}/{wb_proj_name}/{search_id}")]

    log(f"Returned {len(runs)} runs")

    ret_data = []
    for rank_i, run in enumerate(runs):
        dat = {"rank": rank_i}
        for f in select_fields:
            v = None
            if f == "last_model":

                parts = proj_cfg["ckpt_cfg_key"].split(".")
                model_path = run.config
                for k in parts:
                    model_path = model_path[k]
                if proj_cfg.get("ckpt_append_name", False):
                    model_path = osp.join(model_path, run.name)

                if not osp.exists(model_path):
                    raise ValueError(f"Could not locate model folder {model_path}")
                model_idxs = [
                    int(model_f.split("ckpt.")[1].split(".pth")[0])
                    for model_f in os.listdir(model_path)
                    if model_f.startswith("ckpt.")
                ]
                if len(model_idxs) == 0:
                    raise ValueError(f"No models found under {model_path}")
                max_idx = max(model_idxs)
                final_model_f = osp.join(model_path, f"ckpt.{max_idx}.pth")
                v = final_model_f
            elif f == "summary":
                v = dict(run.summary)
                v["status"] = str(run.state)
                # Filter out non-primitive values.
                v = {
                    k: k_v for k, k_v in v.items() if isinstance(k_v, (int, float, str))
                }

            elif f == "status":
                v = run.state
            elif f == "config":
                v = run.config
            elif f == "id":
                v = run.id
            elif f.startswith("config."):
                config_parts = f.split("config.")
                parts = config_parts[1].split(".")
                v = run.config
                for k in parts:
                    v = v[k]
            else:
                if f.startswith("ALL_"):
                    fetch_field = extract_query_key(f)
                    df = run.history(samples=15000)
                    if fetch_field not in df.columns:
                        raise ValueError(
                            f"Could not find {fetch_field} in {df.columns} for query {filter_fields}"
                        )
                    v = df[["_step", fetch_field]]
                else:
                    if f not in run.summary:
                        if error_ok:
                            continue
                        raise ValueError(
                            f"Could not find {f} in {run.summary.keys()} from run {run} with query {query_dict}"
                        )
                    v = run.summary[f]
            if v is not None:
                dat[f] = v
        if len(dat) > 0:
            ret_data.append(dat)
        if limit is not None and len(ret_data) >= limit:
            break

    cache.save(ret_data)
    if reduce_op is not None:
        reduce_data = defaultdict(list)
        for p in ret_data:
            for k, v in p.items():
                reduce_data[k].append(v)
        ret_data = {k: reduce_op(v) for k, v in reduce_data.items()}

    log(f"Got data {ret_data}")
    return ret_data


def query_s(
    query_str: str,
    proj_cfg: DictConfig,
    verbose=True,
    use_cached: bool = False,
):

    select_s, filter_s = query_str.split(" WHERE ")
    select_fields = select_s.replace(" ", "").split(",")

    parts = filter_s.split(" LIMIT ")
    filter_s = parts[0]

    limit = None
    if len(parts) > 1:
        limit = int(parts[1])

    filter_fields = filter_s.replace(" ", "").split(",")
    filter_fields = [s.split("=") for s in filter_fields]
    filter_fields = {k: v for k, v in filter_fields}

    return query(
        select_fields,
        filter_fields,
        proj_cfg,
        verbose=verbose,
        limit=limit,
        use_cached=use_cached,
    )


def fetch_data_from_cfg(
    plot_cfg_path: str,
    add_query_fields: Optional[List[str]] = None,
    error_ok: bool = False,
) -> pd.DataFrame:
    """
    See the README for how the YAML file at `plot_cfg_path` should be structured.
    """

    cfg = OmegaConf.load(plot_cfg_path)
    if add_query_fields is None:
        add_query_fields = []

    query_k = cfg.plot_key

    result = batch_query(
        [
            [query_k, *add_query_fields, *cfg.get("add_query_keys", [])]
            for _ in cfg.methods
        ],
        [{cfg.method_spec: v} for v in cfg.methods.values()],
        all_should_skip=[len(v) == 0 for v in cfg.methods.values()],
        all_add_info=[{"method": k} for k in cfg.methods.keys()],
        proj_cfg=OmegaConf.load(cfg.proj_cfg),
        use_cached=cfg.use_cached,
        verbose=False,
        error_ok=error_ok,
    )
    return pd.DataFrame(result)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--cfg", required=True, type=str)
    parser.add_argument("--cache", action="store_true")
    args, query_args = parser.parse_known_args()
    query_args = " ".join(query_args)
    proj_cfg = OmegaConf.load(args.cfg)

    result = query_s(query_args, proj_cfg, use_cached=args.cache, verbose=False)
    result_summary = {}
    keys = list(result[0].keys())
    for k in keys:
        values = [r[k] for r in result]
        if isinstance(values[0], float):
            result_summary[f"{k} (mean, std)"] = (
                np.mean(values),
                np.std(values),
            )

    pprint(result)
    if len(result_summary) > 0:
        pprint(result_summary)



from .pointmass_env import PointMassEnv, PointMassParams
from .pointmass_obstacle import (
    PointMassObstacleEnv,
    PointMassObstacleParams,
    SquareObstacle,
)

__all__ = [
    "PointMassEnv",
    "PointMassParams",
    "PointMassObstacleEnv",
    "PointMassObstacleParams",
    "SquareObstacle",
]



from typing import List

import torch.nn as nn


def make_mlp_layers(
    input_dim: int,
    output_dim: int,
    hidden_dim: int,
    n_layers: int,
    activations_cls=None,
) -> List[nn.Module]:
    """
    :returns: A list of modules in the mlp.
    """
    if activations_cls is None:
        activations_cls = nn.Tanh
    net_layers = [nn.Linear(input_dim, hidden_dim)]
    for _ in range(n_layers):
        net_layers.extend(
            [
                activations_cls(),
                nn.Linear(hidden_dim, hidden_dim),
            ]
        )
    net_layers.extend([activations_cls(), nn.Linear(hidden_dim, output_dim)])
    return net_layers



from functools import partial
from typing import Callable, Optional

import gym
import torch

import rl_utils.envs.pointmass  # noqa: F401
from rl_utils.envs.registry import full_env_registry
from rl_utils.envs.vec_env.dummy_vec_env import DummyVecEnv
from rl_utils.envs.vec_env.shmem_vec_env import ShmemVecEnv
from rl_utils.envs.vec_env.vec_env import VecEnv
from rl_utils.envs.vec_env.vec_env_wrappers import (
    VecEnvClipActions,
    VecEnvPermuteFrames,
)
from rl_utils.envs.vec_env.vec_monitor import VecMonitor
from rl_utils.envs.wrappers import TimeLimitMask, VecPyTorch, VecPyTorchFrameStack

try:
    import habitat.utils.gym_definitions  # noqa: F401
except ImportError:
    pass


def create_vectorized_envs(
    env_id: str,
    num_envs: int,
    seed: int = 0,
    *,
    device: Optional[torch.device] = None,
    context_mode: str = "spawn",
    create_env_fn: Optional[Callable[[int], None]] = None,
    force_multi_proc: bool = False,
    num_frame_stack: Optional[int] = None,
    clip_actions: bool = False,
    permute_frames: bool = False,
    **kwargs,
) -> VecEnv:
    found_full_env_cls = full_env_registry.search_env(env_id)
    if found_full_env_cls is not None:
        # print(f"Found {found_full_env_cls} for env {env_id}")
        return found_full_env_cls(num_envs=num_envs, seed=seed, device=device, **kwargs)

    def full_create_env(rank):
        full_seed = seed + rank
        if create_env_fn is None:
            env = gym.make(env_id)
        else:
            env = create_env_fn(full_seed)
        if str(env.__class__.__name__).find("TimeLimit") >= 0:
            env = TimeLimitMask(env)
        env.seed(full_seed)
        if hasattr(env.action_space, "seed"):
            env.action_space.seed(full_seed)
        return env

    envs = [partial(full_create_env, rank=i) for i in range(num_envs)]

    if num_envs > 1 or force_multi_proc:
        envs = ShmemVecEnv(envs, context=context_mode)
    else:
        envs = DummyVecEnv(envs)

    if device is None:
        device = torch.device("cpu")

    envs = VecMonitor(envs)
    envs = VecPyTorch(envs, device)

    if permute_frames:
        envs = VecEnvPermuteFrames(envs)
    if clip_actions:
        envs = VecEnvClipActions(envs)

    if num_frame_stack is not None:
        return VecPyTorchFrameStack(envs, num_frame_stack, device)
    return envs



"""
Code is from https://github.com/facebookresearch/habitat-lab/blob/main/habitat_baselines/rl/models/simple_cnn.py
"""

from typing import Dict, List, Union

import numpy as np
import torch
from torch import nn as nn


class SimpleCNN(nn.Module):
    """A Simple 3-Conv CNN followed by a fully connected layer

    Takes in observations and produces an embedding of the rgb and/or depth
        components. Note the observation spaces should be in format (W, H, C) but
        the input to the model should be in format (C, W, H).

    :param obs_shape: Either a dictionary of observation space shapes or the
        observation space shape. This will only select the keys from the
        observation space have 3 tensor dimensions.
    :param output_size: The hidden dimension output size
    """

    def __init__(
        self,
        obs_shape: Union[Dict[str, List[int]], List[int]],
        output_size: int,
    ):
        super().__init__()
        if not isinstance(obs_shape, dict):
            obs_shape = {None: obs_shape}

        obs_input_keys = [k for k, v in obs_shape.items() if len(v) == 3]

        self._n_input = sum(obs_shape[k][2] for k in obs_input_keys)
        self._obs_input_keys = obs_input_keys

        # kernel size for different CNN layers
        self._cnn_layers_kernel_size = [(8, 8), (4, 4), (3, 3)]

        # strides for different CNN layers
        self._cnn_layers_stride = [(4, 4), (2, 2), (1, 1)]

        cnn_dims = np.array(obs_shape[obs_input_keys[0]][:2], dtype=np.float32)
        for kernel_size, stride in zip(
            self._cnn_layers_kernel_size, self._cnn_layers_stride
        ):
            cnn_dims = self._conv_output_dim(
                dimension=cnn_dims,
                padding=np.array([0, 0], dtype=np.float32),
                dilation=np.array([1, 1], dtype=np.float32),
                kernel_size=np.array(kernel_size, dtype=np.float32),
                stride=np.array(stride, dtype=np.float32),
            )

        self.cnn = nn.Sequential(
            nn.Conv2d(
                in_channels=self._n_input,
                out_channels=32,
                kernel_size=self._cnn_layers_kernel_size[0],
                stride=self._cnn_layers_stride[0],
            ),
            nn.ReLU(True),
            nn.Conv2d(
                in_channels=32,
                out_channels=64,
                kernel_size=self._cnn_layers_kernel_size[1],
                stride=self._cnn_layers_stride[1],
            ),
            nn.ReLU(True),
            nn.Conv2d(
                in_channels=64,
                out_channels=32,
                kernel_size=self._cnn_layers_kernel_size[2],
                stride=self._cnn_layers_stride[2],
            ),
            #  nn.ReLU(True),
            nn.Flatten(),
            nn.Linear(32 * cnn_dims[0] * cnn_dims[1], output_size),
            nn.ReLU(True),
        )

        self.layer_init()

    def _conv_output_dim(self, dimension, padding, dilation, kernel_size, stride):
        r"""Calculates the output height and width based on the input
        height and width to the convolution layer.

        ref: https://pytorch.org/docs/master/nn.html#torch.nn.Conv2d
        """
        assert len(dimension) == 2
        out_dimension = []
        for i in range(len(dimension)):
            out_dimension.append(
                int(
                    np.floor(
                        (
                            (
                                dimension[i]
                                + 2 * padding[i]
                                - dilation[i] * (kernel_size[i] - 1)
                                - 1
                            )
                            / stride[i]
                        )
                        + 1
                    )
                )
            )
        return tuple(out_dimension)

    def layer_init(self):
        for layer in self.cnn:  # type: ignore
            if isinstance(layer, (nn.Conv2d, nn.Linear)):
                nn.init.kaiming_normal_(layer.weight, nn.init.calculate_gain("relu"))
                if layer.bias is not None:
                    nn.init.constant_(layer.bias, val=0)

    def forward(self, obs: Union[Dict[str, torch.Tensor], torch.Tensor]):
        if isinstance(obs, dict):
            cnn_inputs = torch.cat([obs[k] for k in self._obs_input_keys], dim=1)
        else:
            cnn_inputs = obs
        return self.cnn(cnn_inputs)



import argparse
from typing import Callable, Dict, List, Optional

import pandas as pd
from omegaconf import OmegaConf

from rl_utils.plotting.utils import MISSING_VALUE
from rl_utils.plotting.wb_query import fetch_data_from_cfg


def plot_table(
    df: pd.DataFrame,
    col_key: str,
    row_key: str,
    cell_key: str,
    col_order: List[str],
    row_order: List[str],
    renames: Optional[Dict[str, str]] = None,
    error_scaling=1.0,
    n_decimals=2,
    missing_fill_value=MISSING_VALUE,
    error_fill_value=0.3444,
    get_row_highlight: Optional[Callable[[str, pd.DataFrame], Optional[str]]] = None,
    make_col_header: Optional[Callable[[int], str]] = None,
    x_label: str = "",
    y_label: str = "",
    skip_toprule: bool = False,
    include_err: bool = True,
    write_to=None,
    err_key: Optional[str] = None,
    add_tabular: bool = True,
    bold_row_names: bool = True,
    show_row_labels: bool = True,
    compute_err_fn: Optional[Callable[[pd.Series], pd.Series]] = None,
):
    """
    :param df: The index of the data frame does not matter, only the row values and column names matter.
    :param col_key: A string from the set of columns.
    :param row_key: A string from the set of columns (but this is used to form the rows of the table).
    :param renames: Only used for display name conversions. Does not affect functionality.
    :param make_col_header: Returns the string at the top of the table like
        "ccccc". Put "c|ccccc" to insert a vertical line in between the first
        and other columns.
    :param x_label: Renders another row of text on the top that spans all the columns.
    :param y_label: Renders a side column with vertically rotated text that spawns all the rows.
    :param err_key: If non-None, this will be used as the error and override any error calculation.
    :param show_row_labels: If False, the row names are not diplayed, and no
        column for the row name is displayed.

    Example: the data fame might look like
    ```
       democount        type  final_train_success
    0     100  mirl train               0.9800
    1     100  mirl train               0.9900
    3     100   mirl eval               1.0000
    4     100   mirl eval               1.0000
    12     50  mirl train               0.9700
    13     50  mirl train               1.0000
    15     50   mirl eval               1.0000
    16     50   mirl eval               0.7200
    ```
    `col_key='type', row_key='demcount',
    cell_key='final_train_success'` plots the # of demos as rows and
    the type as columns with the final_train_success values as the cell
    values. Duplicate row and columns are automatically grouped
    together.

    """
    if make_col_header is None:

        def make_col_header(n_cols):
            return "c" * n_cols

    if renames is None:
        renames = {}
    df = df.replace("missing", missing_fill_value)
    df = df.replace("error", error_fill_value)

    rows = {}
    for row_k, row_df in df.groupby(row_key):
        grouped = row_df.groupby(col_key)
        df_avg_y = grouped[cell_key].mean()
        df_std_y = grouped[cell_key].std() * error_scaling

        sel_err = False
        if err_key is not None:
            err = grouped[err_key].mean()
            if not err.hasnans:
                df_std_y = err
                sel_err = True

        if not sel_err and compute_err_fn is not None:
            df_std_y = compute_err_fn(grouped[cell_key])

        rows[row_k] = (df_avg_y, df_std_y)

    col_sep = " & "
    row_sep = " \\\\\n"

    all_s = []

    def clean_text(s):
        return s.replace("%", "\\%").replace("_", " ")

    # Add the column title row.
    row_str = []
    if show_row_labels:
        row_str.append("")
    for col_k in col_order:
        row_str.append("\\textbf{%s}" % clean_text(renames.get(col_k, col_k)))
    all_s.append(col_sep.join(row_str))

    for row_k in row_order:
        if row_k == "hline":
            all_s.append("\\hline")
            continue
        row_str = []

        if show_row_labels:
            if bold_row_names:
                row_str.append("\\textbf{%s}" % clean_text(renames.get(row_k, row_k)))
            else:
                row_str.append(clean_text(renames.get(row_k, row_k)))

        row_y, row_std = rows[row_k]

        if get_row_highlight is not None:
            sel_col = get_row_highlight(row_k, row_y)
        else:
            sel_col = None
        for col_k in col_order:
            if col_k not in row_y:
                row_str.append("-")
            else:

                val = row_y.loc[col_k]
                std = row_std.loc[col_k]
                if val == missing_fill_value:
                    row_str.append("-")
                elif val == error_fill_value:
                    row_str.append("E")
                else:
                    err = ""
                    if include_err:
                        err = f"$ \\pm$ %.{n_decimals}f " % std
                    txt = f" %.{n_decimals}f {{\\scriptsize}}{err} " % val

                    if col_k == sel_col:
                        txt = "\\textbf{ " + txt + " }"
                    row_str.append(txt)

        all_s.append(col_sep.join(row_str))

    n_columns = len(col_order)
    if show_row_labels:
        n_columns += 1
    col_header_s = make_col_header(n_columns)
    if y_label != "":
        col_header_s = "c" + col_header_s
        start_of_line = " & "
        toprule = ""

        midrule = "\\cmidrule{2-%s}\n" % (n_columns + 1)
        botrule = midrule
        row_lines = [start_of_line + x for x in all_s[1:]]
        row_lines[0] = (
            "\\multirow{4}{1em}{\\rotatebox{90}{%s}}" % y_label
        ) + row_lines[0]
    else:
        row_lines = all_s[1:]
        start_of_line = ""
        toprule = "\\toprule\n"
        midrule = "\\midrule\n"
        botrule = "\\bottomrule"

    if skip_toprule:
        toprule = ""

    if x_label != "":
        toprule += ("& \\multicolumn{%i}{c}{%s}" % (n_columns, x_label)) + row_sep

    ret_s = ""
    if add_tabular:
        ret_s += "\\begin{tabular}{%s}\n" % col_header_s
        # Line above the table.
        ret_s += toprule

        # Separate the column headers from the rest of the table by a line.
        ret_s += start_of_line + all_s[0] + row_sep
        ret_s += midrule

    all_row_s = ""
    for row_line in row_lines:
        all_row_s += row_line

        # Do not add the separator to the last element if we are not in tabular mode.
        if "hline" not in row_line:
            all_row_s += row_sep
        else:
            all_row_s += "\n"

    ret_s += all_row_s
    # Line below the table.
    if add_tabular:
        ret_s += botrule

        ret_s += "\n\\end{tabular}\n"

    if write_to is not None:
        with open(write_to, "w") as f:
            f.write(ret_s)
        print(f"Wrote result to {write_to}")
    else:
        print(ret_s)

    return ret_s


def plot_from_file(plot_cfg_path, add_query_fields=None):
    cfg = OmegaConf.load(plot_cfg_path)
    df = fetch_data_from_cfg(plot_cfg_path, add_query_fields)

    plot_table(df, cell_key=cfg.plot_key, **cfg.sub_plot_params)
    return df


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cfg", type=str, required=True)
    args = parser.parse_args()
    plot_from_file(args.cfg)



import torch


def cumsum_discount(x: torch.Tensor, discount: float) -> torch.Tensor:
    """
    :param x: Tensor of shape (batch_size, ..., 1)
    :returns: Tensor of shape same size as `x`.
    """

    r = torch.zeros_like(x)
    r[-1] = x[-1]
    for t in reversed(range(x.shape[0] - 1)):
        r[t] = x[t] + discount * r[t + 1]
    return r



import os
import os.path as osp

import cv2
import numpy as np


def save_mp4(frames, vid_dir: str, name: str, fps: int = 60, should_print=True):
    """
    :param name: The name WITHOUT the ".mp4" extension.
    """
    frames = np.array(frames)
    if len(frames[0].shape) == 4:
        new_frames = frames[0]
        for i in range(len(frames) - 1):
            new_frames = np.concatenate([new_frames, frames[i + 1]])
        frames = new_frames

    if not osp.exists(vid_dir):
        os.makedirs(vid_dir)

    vid_file = osp.join(vid_dir, name + ".mp4")
    if osp.exists(vid_file):
        os.remove(vid_file)

    w, h = frames[0].shape[:-1]
    videodims = (h, w)
    fourcc = cv2.VideoWriter_fourcc("m", "p", "4", "v")
    video = cv2.VideoWriter(vid_file, fourcc, fps, videodims)
    for frame in frames:
        frame = frame[..., 0:3][..., ::-1]
        video.write(frame)
    video.release()
    if should_print:
        print(f"Rendered to {vid_file}")



import contextlib
import os
from abc import ABC, abstractmethod

import cloudpickle

from rl_utils.common.tile_images import tile_images

FINAL_OBS_KEY = "final_obs"


class AlreadySteppingError(Exception):
    """
    Raised when an asynchronous step is running while
    step_async() is called again.
    """

    def __init__(self):
        msg = "already running an async step"
        Exception.__init__(self, msg)


class NotSteppingError(Exception):
    """
    Raised when an asynchronous step is not running but
    step_wait() is called.
    """

    def __init__(self):
        msg = "not running an async step"
        Exception.__init__(self, msg)


class VecEnv(ABC):
    """
    An abstract asynchronous, vectorized environment.
    Used to batch data from multiple copies of an environment, so that
    each observation becomes an batch of observations, and expected action is a batch of actions to
    be applied per-environment.
    """

    closed = False
    viewer = None

    metadata = {"render.modes": ["human", "rgb_array"]}

    def __init__(self, num_envs, observation_space, action_space):
        self.num_envs = num_envs
        self.observation_space = observation_space
        self.action_space = action_space

    @abstractmethod
    def reset(self):
        """
        Reset all the environments and return an array of
        observations, or a dict of observation arrays.

        If step_async is still doing work, that work will
        be cancelled and step_wait() should not be called
        until step_async() is invoked again.
        """

    @abstractmethod
    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.

        You should not call this if a step_async run is
        already pending.
        """

    @abstractmethod
    def step_wait(self):
        """
        Wait for the step taken with step_async().

        Returns (obs, rews, dones, infos):
         - obs: an array of observations, or a dict of
                arrays of observations.
         - rews: an array of rewards
         - dones: an array of "episode done" booleans
         - infos: a sequence of info objects
        """

    def close_extras(self):
        """
        Clean up the  extra resources, beyond what's in this base class.
        Only runs when not self.closed.
        """

    def close(self):
        if self.closed:
            return
        if self.viewer is not None:
            self.viewer.close()
        self.close_extras()
        self.closed = True

    def step(self, actions):
        """
        Step the environments synchronously.

        This is available for backwards compatibility.
        """
        self.step_async(actions)
        return self.step_wait()

    def render(self, mode="human", **kwargs):
        imgs = self.get_images(mode=mode, **kwargs)
        bigimg = tile_images(imgs)
        if mode == "human":
            self.get_viewer().imshow(bigimg)
            return self.get_viewer().isopen
        else:
            return bigimg

    def get_images(self, mode=None, **kwargs):
        """
        Return RGB images from each environment
        """
        raise NotImplementedError

    @property
    def unwrapped(self):
        if isinstance(self, VecEnvWrapper):
            return self.venv.unwrapped
        else:
            return self

    def get_viewer(self):
        if self.viewer is None:
            from gym.envs.classic_control import rendering

            self.viewer = rendering.SimpleImageViewer()
        return self.viewer


class VecEnvWrapper(VecEnv):
    """
    An environment wrapper that applies to an entire batch
    of environments at once.
    """

    def __init__(self, venv, observation_space=None, action_space=None):
        self.venv = venv
        VecEnv.__init__(
            self,
            num_envs=venv.num_envs,
            observation_space=observation_space or venv.observation_space,
            action_space=action_space or venv.action_space,
        )

    def step_async(self, actions):
        self.venv.step_async(actions)

    def reset(self):
        return self.venv.reset()

    def step_wait(self):
        return self.venv.step_wait()

    def close(self):
        return self.venv.close()

    def render(self, mode="human", **kwargs):
        return self.venv.render(mode=mode, **kwargs)

    def get_images(self, mode=None, **kwargs):
        return self.venv.get_images(mode=mode, **kwargs)


class VecEnvObservationWrapper(VecEnvWrapper):
    @abstractmethod
    def process(self, obs):
        pass

    def reset(self):
        obs = self.venv.reset()
        return self.process(obs)

    def step_wait(self):
        obs, rews, dones, infos = self.venv.step_wait()
        return self.process(obs), rews, dones, infos


class CloudpickleWrapper:
    """
    Uses cloudpickle to serialize contents
    """

    def __init__(self, x):
        self.x = x

    def __getstate__(self):
        return cloudpickle.dumps(self.x)

    def __setstate__(self, ob):
        self.x = cloudpickle.loads(ob)


@contextlib.contextmanager
def clear_mpi_env_vars():
    """
    from mpi4py import MPI will call MPI_Init by default.  If the child process has MPI environment variables, MPI will think that the child process is an MPI process just like the parent and do bad things such as hang.
    This context manager is a hacky way to clear those environment variables temporarily such as when we are starting multiprocessing
    Processes.
    """
    removed_environment = {}
    for k, v in list(os.environ.items()):
        for prefix in ["OMPI_", "PMI_"]:
            if k.startswith(prefix):
                removed_environment[k] = v
                del os.environ[k]
    try:
        yield
    finally:
        os.environ.update(removed_environment)



import argparse
import os
import os.path as osp
import random
import string
import uuid

try:
    import libtmux
except ImportError:
    libtmux = None
import shlex

from omegaconf import OmegaConf

from rl_utils.plotting.wb_query import query_s


def get_arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--sess-id",
        type=int,
        default=-1,
        help="tmux session id to connect to. If unspec will run in current window",
    )
    parser.add_argument(
        "--sess-name",
        default=None,
        type=str,
        help="tmux session name to connect to",
    )
    parser.add_argument("--proj-dat", type=str, default=None)
    parser.add_argument("--conda-env", type=str, default=None)
    parser.add_argument("--runs-dir", type=str, default="data/log/runs")
    parser.add_argument(
        "--group-id",
        type=str,
        default=None,
        help="If not assigned then a randomly assigned one is generated.",
    )
    parser.add_argument(
        "--name-prefix",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--run-single",
        action="store_true",
        help="""
            If true, will run all commands in a single pane sequentially. This
            will chain together multiple runs in a cmd file rather than run
            them sequentially.
    """,
    )
    parser.add_argument(
        "--cd",
        default="-1",
        type=str,
        help="""
            String of CUDA_VISIBLE_DEVICES. A value of "-1" will not set
            CUDA_VISIBLE_DEVICES at all.
            """,
    )

    parser.add_argument(
        "--skip-add-all",
        action="store_true",
    )

    parser.add_argument("--cfg", type=str, default=None)
    parser.add_argument("--base-data-dir", type=str, default=None)

    # MULTIPROC OPTIONS
    parser.add_argument("--pt-proc", type=int, default=-1)

    # SLURM OPTIONS
    parser.add_argument("--comment", type=str, default=None)
    parser.add_argument(
        "--slurm",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--slurm-no-batch",
        action="store_true",
        help="""
            If specified, will run with srun instead of sbatch
        """,
    )
    parser.add_argument(
        "--skip-env",
        action="store_true",
        help="""
            If true, will not export any environment variables from config.yaml
            """,
    )
    parser.add_argument(
        "--speed",
        action="store_true",
        help="""
            SLURM optimized for maximum CPU usage.
            """,
    )
    parser.add_argument(
        "--partition", type=str, default=None, help="Slum parition type."
    )
    parser.add_argument(
        "--time",
        type=str,
        default=None,
        help="""
            Slurm time limit. "10:00" is 10 minutes.
            """,
    )
    parser.add_argument(
        "--c",
        type=str,
        default=None,
        help="""
            Number of cpus for SLURM job
            """,
    )
    parser.add_argument(
        "--constraint",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--cpu-mem",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--g",
        type=str,
        default="1",
        help="""
            Number of gpus for SLURM job
            """,
    )
    parser.add_argument(
        "--ntasks",
        type=str,
        default=None,
        help="""
            Number of processes for SLURM job
            """,
    )

    return parser


def add_on_args(spec_args):
    spec_args = ['"' + x + '"' if " " in x else x for x in spec_args]
    return " ".join(spec_args)


def get_cmds(rest):
    cmd = rest[0]
    if len(rest) > 1:
        rest = " ".join(rest[1:])
    else:
        rest = ""

    if ".cmd" in cmd:
        with open(cmd) as f:
            cmds = f.readlines()
    else:
        cmds = [cmd]

    cmds = list(filter(lambda x: not (x.startswith("#") or x == "\n"), cmds))
    return [f"{cmd.rstrip()} {rest}" for cmd in cmds]


def get_tmux_window(sess_name, sess_id):
    if libtmux is None:
        raise ValueError("Must install libtmux to use auto tmux capability")
    server = libtmux.Server()

    if sess_name is None:
        sess = server.get_by_id("$%i" % sess_id)
    else:
        sess = server.find_where({"session_name": sess_name})
    if sess is None:
        raise ValueError("invalid session id")

    return sess.new_window(attach=False, window_name="auto_proc")


def as_list(x, max_num):
    if isinstance(x, int):
        return [x for _ in range(max_num)]
    x = x.split("|")
    if len(x) == 1:
        return [x[0] for _ in range(max_num)]
    return x


def get_random_id() -> str:
    rnd_id = str(uuid.uuid4())[:8]
    return random.choice(string.ascii_uppercase) + rnd_id


def get_cmd_run_str(cmd, args, cmd_idx, num_cmds, proj_cfg):
    conda_env = proj_cfg["conda_env"]

    if args.conda_env is None:
        python_path = osp.join(
            osp.expanduser("~"), "miniconda3", "envs", conda_env, "bin"
        )
        python_path = proj_cfg.get("conda_path", python_path)
    else:
        python_path = args.conda_env

    g = as_list(args.g, num_cmds)
    c = as_list(args.c, num_cmds)
    if args.ntasks is None:
        ntasks = g[:]
    else:
        ntasks = as_list(args.ntasks, num_cmds)
    ident = get_random_id()
    if args.name_prefix is not None:
        ident = args.name_prefix + "_" + ident
    log_file = osp.join(args.runs_dir, ident) + ".log"
    cmd = cmd.replace("$SLURM_ID", ident)

    if args.partition is None:
        env_vars = " ".join(proj_cfg["add_env_vars"])
        return f"{env_vars} {cmd}"
    else:

        if not args.slurm_no_batch:
            run_file, run_name = generate_slurm_batch_file(
                log_file,
                ident,
                python_path,
                cmd,
                args.partition,
                ntasks[cmd_idx],
                g[cmd_idx],
                c[cmd_idx],
                args,
                proj_cfg,
            )
            return f"sbatch {run_file}"
        else:
            srun_settings = (
                f"--gres=gpu:{args.g} "
                + f"-p {args.partition} "
                + f"-c {args.c} "
                + f"-J {ident} "
                + f"-o {log_file}"
            )

            # This assumes the command begins with "python ..."
            return f"srun {srun_settings} {python_path}/{cmd}"


def sub_wb_query(cmd, proj_cfg):
    parts = cmd.split("&")
    if len(parts) < 3:
        return [cmd]

    new_cmd = [parts[0]]
    parts = parts[1:]

    for i in range(len(parts)):
        if i % 2 == 0:
            wb_query = parts[i]
            result = query_s(wb_query, proj_cfg, verbose=False)
            if len(result) == 0:
                raise ValueError(f"Got no response from {wb_query}")
            sub_vals = []
            for match in result:
                del match["rank"]
                if len(match) > 1:
                    raise ValueError(f"Only single value query supported, got {match}")
                sub_val = list(match.values())[0]
                sub_vals.append(sub_val)

            new_cmd = [c + sub_val for c in new_cmd for sub_val in sub_vals]
        else:
            for j in range(len(new_cmd)):
                new_cmd[j] += parts[i]

    return new_cmd


def log(s, args):
    print(s)


def split_cmd(cmd):
    cmd_parts = cmd.split(" ")
    ret_cmds = [[]]
    for cmd_part in cmd_parts:
        prefix = ""
        if "=" in cmd_part:
            prefix, cmd_part = cmd_part.split("=")
            prefix += "="

        if "," in cmd_part:
            ret_cmds = [
                ret_cmd + [prefix + split_part]
                for ret_cmd in ret_cmds
                for split_part in cmd_part.split(",")
            ]
        else:
            ret_cmds = [ret_cmd + [prefix + cmd_part] for ret_cmd in ret_cmds]
    return [" ".join(ret_cmd) for ret_cmd in ret_cmds]


def sub_in_args(old_cmd: str, new_args: str):
    old_parts = shlex.split(old_cmd, posix=False)
    new_parts = shlex.split(new_args, posix=False)

    i = 0
    while i < len(new_parts):
        if new_parts[i] in old_parts:
            old_i = old_parts.index(new_parts[i])
            old_parts[old_i + 1] = new_parts[i + 1]
        else:
            old_parts.extend(new_parts[i : i + 2])
        i += 2
    return " ".join(old_parts)


def execute_command_file(run_cmd, args, proj_cfg):
    if not osp.exists(args.runs_dir):
        os.makedirs(args.runs_dir)

    cmds = get_cmds(run_cmd)

    # Sub in W&B args
    cmds = [c for cmd in cmds for c in sub_wb_query(cmd, proj_cfg)]

    # Split the commands.
    cmds = [c for cmd in cmds for c in split_cmd(cmd)]

    n_cmds = len(cmds)

    add_all = proj_cfg.get("add_all", None)
    if add_all is not None and not args.skip_add_all:
        cmds = [sub_in_args(cmd, add_all) for cmd in cmds]

    # Add on the project data
    if args.proj_dat is not None:
        proj_data = proj_cfg.get("proj_data", {})
        for k in args.proj_dat.split(","):
            cmds = [sub_in_args(cmd, proj_data[k]) for cmd in cmds]
            env_var_dat = proj_cfg.get("proj_dat_add_env_vars", {}).get(k, None)
            if env_var_dat is not None:
                cmds = [env_var_dat + " " + cmd for cmd in cmds]

    # Sub in variables
    if args.base_data_dir is not None:
        cmds = [cmd.replace("$DATA_DIR", args.base_data_dir) for cmd in cmds]
    elif "base_data_dir" in proj_cfg:
        cmds = [cmd.replace("$DATA_DIR", proj_cfg["base_data_dir"]) for cmd in cmds]

    if args.group_id is None:
        group_ident = get_random_id()
    else:
        group_ident = args.group_id
    print(f"Assigning group ID {group_ident}")
    cmds = [cmd.replace("$GROUP_ID", group_ident) for cmd in cmds]
    cmds = [cmd.replace("$CMD_RANK", str(rank_i)) for rank_i, cmd in enumerate(cmds)]
    cmds = [cmd.replace("$PROJECT_NAME", proj_cfg.proj_name) for cmd in cmds]
    cmds = [cmd.replace("$WB_ENTITY", proj_cfg.wb_entity) for cmd in cmds]

    if args.pt_proc != -1:
        pt_dist_str = f"torchrun --nproc_per_node {args.pt_proc}"

        def make_dist_cmd(x):
            return x.replace("python", pt_dist_str)

        cmds[0] = make_dist_cmd(cmds[0])

    DELIM = " ; "

    cd = as_list(args.cd, n_cmds)

    if args.sess_id == -1 and args.sess_name is None:
        if args.partition is not None:
            for cmd_idx, cmd in enumerate(cmds):
                run_cmd = get_cmd_run_str(cmd, args, cmd_idx, n_cmds, proj_cfg)
                log(f"Running {run_cmd}", args)
                os.system(run_cmd)
        elif args.run_single:
            cmds = [get_cmd_run_str(x, args, 0, 1, proj_cfg) for x in cmds]
            exec_cmd = DELIM.join(cmds)

            log(f"Running {exec_cmd}", args)
            os.system(exec_cmd)

        elif n_cmds == 1:
            exec_cmd = get_cmd_run_str(cmds[0], args, 0, n_cmds, proj_cfg)
            if cd[0] != "-1":
                exec_cmd = "CUDA_VISIBLE_DEVICES=" + cd[0] + " " + exec_cmd
            log(f"Running {exec_cmd}", args)
            os.system(exec_cmd)
        else:
            raise ValueError(
                f"Running multiple jobs. You must specify tmux session id. Tried to run {cmds}"
            )
    else:

        if args.run_single:
            cmds = DELIM.join(cmds)
            cmds = [cmds]

        for cmd_idx, cmd in enumerate(cmds):
            new_window = get_tmux_window(args.sess_name, args.sess_id)

            log("running full command %s\n" % cmd, args)

            run_cmd = get_cmd_run_str(cmd, args, cmd_idx, n_cmds, proj_cfg)

            # Send the keys to run the command
            if args.partition is None:
                last_pane = new_window.attached_pane
                last_pane.send_keys(run_cmd, enter=False)
                pane = new_window.split_window(attach=False)
                pane.set_height(height=50)
                pane.send_keys("source deactivate")

                if "conda_env" in proj_cfg:
                    pane.send_keys("source activate " + proj_cfg["conda_env"])
                pane.enter()
                if cd[cmd_idx] != "-1":
                    pane.send_keys("export CUDA_VISIBLE_DEVICES=" + cd[cmd_idx])
                    pane.enter()
                else:
                    pane.send_keys(run_cmd)

                pane.enter()
            else:
                pane = new_window.split_window(attach=False)
                pane.set_height(height=10)
                pane.send_keys(run_cmd)

        log("everything should be running...", args)


def generate_slurm_batch_file(
    log_file, ident, python_path, cmd, partition, ntasks, g, c, args, proj_cfg
):
    ignore_nodes_s = ",".join(proj_cfg.get("slurm_ignore_nodes", []))
    if len(ignore_nodes_s) != 0:
        ignore_nodes_s = "#SBATCH -x " + ignore_nodes_s

    add_options = [ignore_nodes_s]
    if args.time is not None:
        add_options.append(f"#SBATCH --time={args.time}")
    if args.comment is not None:
        add_options.append(f'#SBATCH --comment="{args.comment}"')
    if args.constraint is not None:
        add_options.append(f"#SBATCH --constraint={args.constraint}")
    if args.cpu_mem is not None:
        add_options.append(f"#SBATCH --mem-per-cpu={args.cpu_mem}")
    add_options = "\n".join(add_options)

    python_parts = cmd.split("python")
    has_python = False
    if len(python_parts) > 1:
        cmd = "python" + python_parts[1]
        has_python = True

    if not args.skip_env:
        env_vars = proj_cfg.get("add_env_vars", [])
        env_vars = [f"export {x}" for x in env_vars]

        if args.proj_dat is not None:
            for k in args.proj_dat.split(","):
                env_var_dat = proj_cfg.get("proj_dat_add_env_vars", {}).get(k, None)
                if env_var_dat is not None:
                    proj_env_vars = env_var_dat.split(" ")
                    for proj_env_var in proj_env_vars:
                        env_vars.append(f"export {proj_env_var}")

        env_vars = "\n".join(env_vars)

    cpu_options = "#SBATCH --cpus-per-task %i" % int(c)
    if args.speed:
        cpu_options = "#SBATCH --overcommit\n"
        cpu_options += "#SBATCH --cpu-freq=performance\n"

    if has_python:
        run_cmd = python_path + "/" + cmd
        requeue_s = "#SBATCH --requeue"
    else:
        run_cmd = cmd
        requeue_s = ""

    fcontents = """#!/bin/bash
#SBATCH --job-name=%s
#SBATCH --output=%s
#SBATCH --gres gpu:%i
%s
#SBATCH --nodes 1
#SBATCH --signal=USR1@600
#SBATCH --ntasks-per-node %i
%s
#SBATCH -p %s
%s

MAIN_ADDR=$(scontrol show hostnames "${SLURM_JOB_NODELIST}" | head -n 1)
export MAIN_ADDR
%s

set -x
srun %s"""
    job_name = ident
    log_file_loc = "/".join(log_file.split("/")[:-1])
    fcontents = fcontents % (
        job_name,
        log_file,
        int(g),
        cpu_options,
        int(ntasks),
        requeue_s,
        partition,
        add_options,
        env_vars,
        run_cmd,
    )
    job_file = osp.join(log_file_loc, job_name + ".sh")
    with open(job_file, "w") as f:
        f.write(fcontents)
    return job_file, job_name


def full_execute_command_file():
    parser = get_arg_parser()
    args, rest = parser.parse_known_args()
    if args.cfg is None:
        proj_cfg = {}
    else:
        proj_cfg = OmegaConf.load(args.cfg)
    slurm_cfg = proj_cfg.get("slurm", {})
    if args.slurm is not None:
        def_slurm = slurm_cfg[args.slurm]
        if args.c is None and "c" in def_slurm:
            args.c = def_slurm["c"]

        if args.time is None and "time" in def_slurm:
            args.time = def_slurm["time"]

        if args.partition is None and "partition" in def_slurm:
            args.partition = def_slurm["partition"]

        if args.constraint is None and "constraint" in def_slurm:
            args.constraint = def_slurm["constraint"]

        if args.cpu_mem is None and "cpu_mem" in def_slurm:
            args.cpu_mem = def_slurm["cpu_mem"]

    if args.c is None:
        args.c = "7"

    execute_command_file(rest, args, proj_cfg)



import argparse
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from omegaconf import OmegaConf

from rl_utils.plotting.utils import combine_dicts_to_df, fig_save
from rl_utils.plotting.wb_query import batch_query

MARKER_ORDER = ["^", "<", "v", "d", "s", "x", "o", ">"]


def smooth_arr(scalars: List[float], weight: float) -> List[float]:
    """
    Taken from the answer here https://stackoverflow.com/questions/42281844/what-is-the-mathematics-behind-the-smoothing-parameter-in-tensorboards-scalar

    :param weight: Between 0 and 1.
    """
    last = scalars[0]  # First value in the plot (first timestep)
    smoothed = []
    for point in scalars:
        smoothed_val = last * weight + (1 - weight) * point  # Calculate smoothed value
        smoothed.append(smoothed_val)  # Save it
        last = smoothed_val  # Anchor the last smoothed value

    return smoothed


def make_steps_match(plot_df, group_key, x_name):
    all_dfs = []
    for _, method_df in plot_df.groupby([group_key]):
        grouped_runs = method_df.groupby(["run"])
        max_len = -1
        max_step_idxs = None
        for _, run_df in grouped_runs:
            if len(run_df) > max_len:
                max_len = len(run_df)
                max_step_idxs = run_df[x_name]
        for _, run_df in grouped_runs:
            run_df[x_name] = max_step_idxs[: len(run_df)]
            all_dfs.append(run_df)
    return pd.concat(all_dfs)


def line_plot(
    plot_df,
    x_name: str,
    y_name: str,
    avg_key: str,
    group_key: str,
    smooth_factor: Union[Dict[str, float], float] = 0.0,
    ax: Optional[matplotlib.axes.Axes] = None,
    y_bounds: Optional[Tuple[float, float]] = None,
    y_disp_bounds: Optional[Tuple[float, float]] = None,
    x_disp_bounds: Optional[Tuple[float, float]] = None,
    group_colors: Optional[Dict[str, int]] = None,
    xtick_fn=None,
    ytick_fn=None,
    legend: bool = False,
    rename_map: Optional[Dict[str, str]] = None,
    title=None,
    axes_font_size=14,
    title_font_size=18,
    legend_font_size="x-large",
    method_idxs: Optional[Dict[str, int]] = None,
    num_marker_points: Optional[Dict[str, int]] = None,
    line_styles: Optional[Dict[str, str]] = None,
    tight=False,
    nlegend_cols=1,
    fetch_std=False,
    y_logscale=False,
    x_logscale=False,
    legend_loc: Optional[str] = None,
    ax_dims: Tuple[int, int] = (5, 4),
):
    """
    :param avg_key: This is typically the seed.
    :param group_key: These are the different lines.
    :param smooth_factor: Can specify a different smooth factor per method if desired.
    :param y_bounds: What the data plot values are clipped to.
    :param y_disp_bounds: What the plotting is stopped at.
    :param ax: If not specified, one is automatically created, with the specified dimensions under `ax_dims`
    :param group_colors: If not specified defaults to `method_idxs`.
    :param num_marker_points: Key maps method name to the number of markers
        drawn on the line, NOT the number of points that are plotted! By
        default this is 8.
    :param legend: Whether to include a legend within the plot.

    :returns: The plotted figure.
    """
    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=ax_dims)
    if rename_map is None:
        rename_map = {}
    if line_styles is None:
        line_styles = {}
    if num_marker_points is None:
        num_marker_points = {}

    if method_idxs is None:
        method_idxs = {k: i for i, k in enumerate(plot_df[group_key].unique())}

    plot_df = plot_df.copy()
    if tight:
        plt.tight_layout(pad=2.2)
    if group_colors is None:
        group_colors = method_idxs

    colors = sns.color_palette()
    group_colors = {k: colors[i] for k, i in group_colors.items()}

    avg_y_df = plot_df.groupby([group_key, x_name]).mean()
    std_y_df = plot_df.groupby([group_key, x_name]).std()

    if y_name in plot_df.columns and y_name not in avg_y_df.columns:
        raise ValueError(
            f"Desired column {y_name} lost in the grouping. Make sure it is a numeric type"
        )

    method_runs = plot_df.groupby(group_key)[avg_key].unique()
    if fetch_std:
        y_std = y_name + "_std"
        new_df = []
        for k, sub_df in plot_df.groupby([group_key]):
            where_matches = avg_y_df.index.get_level_values(0) == k

            use_df = avg_y_df[where_matches]
            if np.isnan(sub_df.iloc[0][y_std]):
                use_df["std"] = std_y_df[where_matches][y_name]
            else:
                use_df["std"] = avg_y_df[where_matches][y_std]
            new_df.append(use_df)
        avg_y_df = pd.concat(new_df)
    else:
        avg_y_df["std"] = std_y_df[y_name]

    lines = []
    names = []

    # Update the legend info with any previously plotted lines
    if ax.get_legend() is not None:
        all_lines = ax.get_lines()

        for i, n in enumerate(ax.get_legend().get_texts()):
            names.append(n.get_text())
            lines.append((all_lines[i * 2 + 1], all_lines[i * 2]))

    if not isinstance(smooth_factor, dict):
        smooth_factor_lookup = defaultdict(lambda: smooth_factor)
    else:
        smooth_factor_lookup = defaultdict(lambda: 0.0)
        for k, v in smooth_factor.items():
            smooth_factor_lookup[k] = v

    for name, sub_df in avg_y_df.groupby(level=0):
        names.append(name)
        x_vals = sub_df.index.get_level_values(x_name).to_numpy()
        y_vals = sub_df[y_name].to_numpy()

        if x_disp_bounds is not None:
            use_y_vals = sub_df[
                sub_df.index.get_level_values(x_name) < x_disp_bounds[1]
            ][y_name].to_numpy()
        else:
            use_y_vals = y_vals
        print(
            f"{name}: n_seeds: {len(method_runs[name])} (from WB run IDs {list(method_runs[name])})",
            max(use_y_vals),
            use_y_vals[-1],
        )
        y_std = sub_df["std"].fillna(0).to_numpy()

        use_smooth_factor = smooth_factor_lookup[name]
        if use_smooth_factor != 0.0:
            y_vals = np.array(smooth_arr(y_vals, use_smooth_factor))
            y_std = np.array(smooth_arr(y_std, use_smooth_factor))

        add_kwargs = {}
        if name in line_styles:
            add_kwargs["linestyle"] = line_styles[name]
        line_to_add = ax.plot(x_vals, y_vals, **add_kwargs)
        sel_vals = [
            int(x)
            for x in np.linspace(0, len(x_vals) - 1, num=num_marker_points.get(name, 8))
        ]
        midx = method_idxs[name] % len(MARKER_ORDER)
        ladd = ax.plot(
            x_vals[sel_vals],
            y_vals[sel_vals],
            MARKER_ORDER[midx],
            label=rename_map.get(name, name),
            color=group_colors[name],
            markersize=8,
        )

        lines.append((ladd[0], line_to_add[0]))

        plt.setp(line_to_add, linewidth=2, color=group_colors[name])
        min_y_fill = y_vals - y_std
        max_y_fill = y_vals + y_std

        if y_bounds is not None:
            min_y_fill = np.clip(min_y_fill, y_bounds[0], y_bounds[1])
            max_y_fill = np.clip(max_y_fill, y_bounds[0], y_bounds[1])

        ax.fill_between(
            x_vals, min_y_fill, max_y_fill, alpha=0.2, color=group_colors[name]
        )
    if y_disp_bounds is not None:
        ax.set_ylim(*y_disp_bounds)
    if x_disp_bounds is not None:
        ax.set_xlim(*x_disp_bounds)

    if xtick_fn is not None:
        plt.xticks(ax.get_xticks(), [xtick_fn(t) for t in ax.get_xticks()])
    if ytick_fn is not None:
        plt.yticks(ax.get_yticks(), [ytick_fn(t) for t in ax.get_yticks()])

    if legend:
        labs = [(i, line_to_add[0].get_label()) for i, line_to_add in enumerate(lines)]
        labs = sorted(labs, key=lambda x: method_idxs[names[x[0]]])
        kwargs = {}
        if legend_loc is not None:
            kwargs["loc"] = legend_loc
        plt.legend(
            [lines[i] for i, _ in labs],
            [x[1] for x in labs],
            fontsize=legend_font_size,
            ncol=nlegend_cols,
            **kwargs,
        )

    ax.grid(b=True, which="major", color="lightgray", linestyle="--")

    ax.set_xlabel(rename_map.get(x_name, x_name), fontsize=axes_font_size)
    ax.set_ylabel(rename_map.get(y_name, y_name), fontsize=axes_font_size)
    if x_logscale:
        ax.set_xscale("log")
    if y_logscale:
        ax.set_yscale("log")
    if title is not None and title != "":
        ax.set_title(title, fontsize=title_font_size)
    return fig, ax


def gen_fake_data(x_scale, data_key, n_runs=5):
    def create_sigmoid():
        noise = np.random.normal(0, 0.01, 100)
        x = np.linspace(0.0, 8.0, 100)
        y = 1 / (1 + np.exp(-x))
        y += noise
        return x, y

    df = None
    for i in range(n_runs):
        x, y = create_sigmoid()
        sub_df = pd.DataFrame({"_step": [int(x_i * x_scale) for x_i in x], data_key: y})
        sub_df["run"] = f"run_{i}"
        if df is None:
            df = sub_df
        else:
            df = pd.concat([df, sub_df])
    df["method"] = "fake"
    return df


def export_legend(ax, line_width, filename):
    fig2 = plt.figure()
    ax2 = fig2.add_subplot(111)
    ax2.axis("off")
    legend = ax2.legend(
        *ax.get_legend_handles_labels(),
        frameon=False,
        loc="lower center",
        ncol=10,
        handlelength=2,
    )
    for line in legend.get_lines():
        line.set_linewidth(line_width)
    fig = legend.figure
    fig.canvas.draw()
    bbox = legend.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(filename, dpi="figure", bbox_inches=bbox)
    print("Saved legend to ", filename)


def plot_legend(
    names: List[str],
    save_path: str,
    plot_colors: Dict[str, int],
    name_map: Optional[Dict[str, str]] = None,
    linestyles: Optional[List[str]] = None,
    darkness: float = 0.1,
    marker_width: float = 0.0,
    marker_size: float = 0.0,
    line_width: float = 3.0,
    alphas: Optional[Dict[str, float]] = None,
):
    """
    :param names: The list of names to appear on the legend.
    :param plot_colors: Maps into the colors of the palette.
    :param name_map: Rename map
    """
    if name_map is None:
        name_map = {}
    if linestyles is None:
        linestyles = []
    if alphas is None:
        alphas = {}

    colors = sns.color_palette()
    group_colors = {name: colors[idx] for name, idx in plot_colors.items()}

    fig, ax = plt.subplots(figsize=(5, 4))
    for name in names:
        add_kwargs = {}
        if name in linestyles:
            linestyle = linestyles[name]
            if isinstance(linestyle, list):
                add_kwargs["linestyle"] = linestyle[0]
                add_kwargs["dashes"] = linestyle[1]
            else:
                add_kwargs["linestyle"] = linestyle

        disp_name = name_map[name]
        midx = plot_colors[name] % len(MARKER_ORDER)
        marker = MARKER_ORDER[midx]
        if marker == "x":
            marker_width = 2.0

        marker_alpha = alphas.get(name, 1.0)
        use_color = (*group_colors[name], marker_alpha)
        ax.plot(
            [0],
            [1],
            marker=marker,
            label=disp_name,
            color=use_color,
            markersize=marker_size,
            markeredgewidth=marker_width,
            # markeredgecolor=(darkness, darkness, darkness, 1),
            markeredgecolor=use_color,
            **add_kwargs,
        )
    export_legend(
        ax,
        line_width,
        save_path,
    )
    plt.clf()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cfg", type=str, required=True)
    args = parser.parse_args()
    cfg = OmegaConf.load(args.cfg)

    query_k = "ALL_" + cfg.plot_key

    result = batch_query(
        [[query_k] for _ in cfg.methods],
        [{cfg.method_spec: v} for v in cfg.methods.values()],
        all_should_skip=[len(v) == 0 for v in cfg.methods.values()],
        all_add_info=[{"method": k} for k in cfg.methods.keys()],
        proj_cfg=OmegaConf.load(cfg.proj_cfg),
        use_cached=cfg.use_cached,
        verbose=False,
    )

    df = combine_dicts_to_df(result, query_k)
    fig = line_plot(df, "_step", cfg.plot_key, "rank", "method", **cfg.plot_params)
    fig_save("data/vis", cfg.save_name, fig)



import os.path as osp
import shutil

import pytest

from rl_utils.common import Evaluator
from rl_utils.envs import create_vectorized_envs
from rl_utils.interfaces import RandomPolicy


@pytest.mark.parametrize(
    "num_render,traj_save_dir",
    [(0, None), (1, None), (None, None), (0, "data/trajs/traj.pt")],
)
def test_reg_eval(num_render, traj_save_dir):
    if osp.exists("data/vids"):
        shutil.rmtree("data/vids")
    if osp.exists("data/trajs"):
        shutil.rmtree("data/trajs")

    envs = create_vectorized_envs(
        "PointMass-v0",
        32,
    )
    evaluator = Evaluator(
        envs, 0, num_render, "data/vids/", fps=10, save_traj_name=traj_save_dir
    )
    policy = RandomPolicy(envs.action_space)
    ret = evaluator.evaluate(policy, 10, 0)
    assert isinstance(ret, dict)
    assert "episode.r" in ret



import argparse
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from omegaconf import OmegaConf

from rl_utils.plotting.utils import fig_save
from rl_utils.plotting.wb_query import batch_query

MISSING_VAL = 0.24444
ERROR_VAL = 0.3444


def plot_bar(
    plot_df,
    group_key,
    plot_key,
    name_ordering=None,
    name_colors=None,
    rename_map=None,
    show_ticks=True,
    axis_font_size=14,
    y_disp_bounds: Tuple[float, float] = None,
    title="",
    error_scaling=1.0,
    missing_fill_value=MISSING_VAL,
    error_fill_value=ERROR_VAL,
):

    def_idx = [(k, i) for i, k in enumerate(plot_df[group_key].unique())]
    if name_ordering is None:
        name_ordering = [x for x, _ in def_idx]
    colors = sns.color_palette()
    if name_colors is None:
        name_colors = {k: colors[v] for k, v in def_idx}
    if rename_map is None:
        rename_map = {}

    plot_df = plot_df.replace("missing", missing_fill_value)
    plot_df = plot_df.replace("error", error_fill_value)
    plot_df[plot_key] = plot_df[plot_key].astype("float")

    df_avg_y = plot_df.groupby(group_key).mean()
    df_std_y = plot_df.groupby(group_key).std()

    avg_y = []
    std_y = []
    name_ordering = [n for n in name_ordering if n in df_avg_y.index]
    is_missing = []
    is_error = []
    for name in name_ordering:
        is_missing.append(df_avg_y[plot_key].loc[name] == missing_fill_value)
        is_error.append(df_avg_y[plot_key].loc[name] == error_fill_value)
        avg_y.append(df_avg_y.loc[name][plot_key])
        std_y.append(df_std_y.loc[name][plot_key] * error_scaling)

    bar_width = 0.35
    bar_darkness = 0.2
    bar_alpha = 0.9
    bar_pad = 0.0
    use_x = np.arange(len(name_ordering))
    colors = [name_colors[x] for x in name_ordering]

    N = len(avg_y)
    start_x = 0.0
    end_x = round(start_x + N * (bar_width + bar_pad), 3)

    use_x = np.linspace(start_x, end_x, N)

    fig, ax = plt.subplots()

    bars = ax.bar(
        use_x,
        avg_y,
        width=bar_width,
        color=colors,
        align="center",
        alpha=bar_alpha,
        yerr=std_y,
        edgecolor=(0, 0, 0, 1.0),
        error_kw={
            "ecolor": (bar_darkness, bar_darkness, bar_darkness, 1.0),
            "lw": 2,
            "capsize": 3,
            "capthick": 2,
        },
    )
    for i, bar in enumerate(bars):
        if is_missing[i]:
            missing_opacity = 0.1
            # prev_color = bar.get_facecolor()
            bar.set_edgecolor((1, 0, 0, missing_opacity))
            bar.set_hatch("//")
        elif is_error[i]:
            missing_opacity = 0.1
            # prev_color = bar.get_facecolor()
            bar.set_edgecolor((0, 0, 1, missing_opacity))
            bar.set_hatch("//")

    if show_ticks:
        xtic_names = [rename_map.get(x, x) for x in name_ordering]
    else:
        xtic_names = ["" for x in name_ordering]

    xtic_locs = use_x
    ax.set_xticks(xtic_locs)
    ax.set_xticklabels(xtic_names, rotation=30)
    ax.set_ylabel(rename_map.get(plot_key, plot_key), fontsize=axis_font_size)
    if y_disp_bounds is not None:
        ax.set_ylim(*y_disp_bounds)
    if title != "":
        ax.set_title(title)
    return fig


def plot_from_file(plot_cfg_path, add_query_fields=None):
    cfg = OmegaConf.load(plot_cfg_path)
    if add_query_fields is None:
        add_query_fields = []

    query_k = cfg.plot_key

    result = batch_query(
        [[query_k, *add_query_fields] for _ in cfg.methods],
        [{cfg.method_spec: v} for v in cfg.methods.values()],
        all_should_skip=[len(v) == 0 for v in cfg.methods.values()],
        all_add_info=[{"method": k} for k in cfg.methods.keys()],
        proj_cfg=OmegaConf.load(cfg.proj_cfg),
        use_cached=cfg.use_cached,
        verbose=False,
    )
    df = pd.DataFrame(result)
    fig = plot_bar(df, "method", query_k, **cfg.plot_params)
    fig_save("data/vis", cfg.save_name, fig)
    return df


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cfg", type=str, required=True)
    args = parser.parse_args()
    plot_from_file(args.cfg)



import torch

from rl_utils.envs import create_vectorized_envs
from rl_utils.envs.pointmass import (
    PointMassObstacleParams,
    PointMassParams,
    SquareObstacle,
)


def test_create():
    envs = create_vectorized_envs(
        "PointMass-v0",
        32,
    )
    envs.reset()
    for _ in range(100):
        rnd_ac = torch.tensor(envs.action_space.sample())
        rnd_ac = rnd_ac.view(1, -1).repeat(32, 1)
        envs.step(rnd_ac)


def test_examples():
    envs = create_vectorized_envs("PointMass-v0", num_envs=32)

    envs = create_vectorized_envs(
        "PointMass-v0",
        num_envs=32,
        params=PointMassParams(dt=0.1, ep_horizon=10),
    )

    envs = create_vectorized_envs("PointMassObstacle-v0", num_envs=32)
    envs.reset()
    for _ in range(100):
        rnd_ac = torch.tensor(envs.action_space.sample())
        rnd_ac = rnd_ac.view(1, -1).repeat(32, 1)
        envs.step(rnd_ac)

    envs = create_vectorized_envs(
        "PointMassObstacle-v0",
        num_envs=32,
        params=PointMassObstacleParams(
            dt=0.1,
            ep_horizon=10,
            square_obstacles=[SquareObstacle((0.5, 0.5), 0.11, 0.5, 45.0)],
        ),
    )
    envs.reset()
    for _ in range(100):
        rnd_ac = torch.tensor(envs.action_space.sample())
        rnd_ac = rnd_ac.view(1, -1).repeat(32, 1)
        envs.step(rnd_ac)



from rl_utils.common.algos import *  # noqa: F403,F401
from rl_utils.common.core_utils import *  # noqa: F403,F401
from rl_utils.common.datasets import *  # noqa: F403,F401
from rl_utils.common.debug_utils import *  # noqa: F403,F401
from rl_utils.common.evaluator import *  # noqa: F403,F401
from rl_utils.common.net_utils import *  # noqa: F403,F401
from rl_utils.common.viz_utils import *  # noqa: F403,F401



import torch

from rl_utils.common import DictDataset, Evaluator, extract_next_tensor
from rl_utils.envs import create_vectorized_envs
from rl_utils.interfaces import RandomPolicy

TRAJ_SAVE_PATH = "data/trajs/traj.pt"


def test_dataset_load_and_manipulation():
    envs = create_vectorized_envs(
        "PointMass-v0",
        32,
    )
    evaluator = Evaluator(
        envs, 0, 0, "data/vids/", fps=10, save_traj_name=TRAJ_SAVE_PATH
    )
    policy = RandomPolicy(envs.action_space)
    evaluator.evaluate(policy, 10, 0)

    dataset = torch.load(TRAJ_SAVE_PATH)
    dataset = extract_next_tensor(dataset)

    dataset = DictDataset(
        dataset,
        ["obs", "actions", "rewards", "masks", "next_obs"],
    )



import numpy as np

from . import VecEnvWrapper


class VecMonitor(VecEnvWrapper):
    def __init__(self, venv):
        VecEnvWrapper.__init__(self, venv)
        self.eprets = None
        self.eplens = None
        self.epcount = 0

    def reset(self):
        self.eprets = np.zeros(self.num_envs, "f")
        self.eplens = np.zeros(self.num_envs, "i")

        return self.venv.reset()

    def step_wait(self):
        obs, rews, dones, infos = self.venv.step_wait()
        self.eprets += rews
        self.eplens += 1
        newinfos = []
        for (i, (done, ret, eplen, info)) in enumerate(
            zip(dones, self.eprets, self.eplens, infos)
        ):
            info = info.copy()
            if done:
                epinfo = {
                    "reward": ret,
                    "length": eplen,
                }
                info["episode"] = epinfo
                self.epcount += 1
                self.eprets[i] = 0
                self.eplens[i] = 0
            newinfos.append(info)

        return obs, rews, dones, newinfos



from rl_utils.logging import Logger


def test_base_logger():
    logger = Logger(
        31,
        "./data/logs",
        "./data/vids",
        "./data/save_dir",
        {},
        5,
        "test_run",
    )
    logger.collect_info("test_val", 2.3)
    logger.collect_info("test_val2", 2.9, no_rolling_window=True)

    logger.interval_log(0, 0)

    logger.close()



from dataclasses import dataclass
from typing import Callable, Optional

import numpy as np
import torch
from gym import spaces
from torch.distributions import Uniform

from rl_utils.envs.registry import full_env_registry
from rl_utils.envs.vec_env.vec_env import FINAL_OBS_KEY, VecEnv


@dataclass(frozen=True)
class PointMassParams:
    """
    :param force_eval_start_dist: Generate the start positions from the eval offset.
    :param force_train_start_dist: Generate the start positions from the train offset.
    :param clip_bounds: Clip the agent to be within [-position_limit, position_limit]^2 ?
    :param clip_actions: Clip the actions to be within -1 to 1.
    :param ep_horizon: The length of the episode.
    :param custom_reward: A function that takes as input the current position,
        previous position, and action  and outputs a reward value. All are PyTorch
        tensors of shape (N,) where N is the number of environments.
    :param random_start_region_sample: If False, then the starting state will
        iterate clockwise around the possible spawning regions.
    """

    force_eval_start_dist: bool = False
    force_train_start_dist: bool = True
    clip_bounds: bool = True
    clip_actions: bool = True
    ep_horizon: int = 5
    num_train_regions: int = 4
    start_state_noise: float = np.pi / 20
    dt: float = 0.2
    reward_dist_pen: float = 1 / 10.0
    start_idx: int = -1
    radius: float = 1.0
    eval_offset: float = 0.0
    train_offset: float = np.pi / 4
    position_limit: float = 1.5
    transition_noise: float = 0.0
    random_start_region_sample: bool = True
    custom_reward: Optional[
        Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]
    ] = None


@full_env_registry.register_env("PointMass-v0")
class PointMassEnv(VecEnv):
    def __init__(
        self,
        num_envs: int,
        params: Optional[PointMassParams] = None,
        device: Optional[torch.device] = None,
        set_eval: bool = False,
        seed: Optional[int] = None,
        **kwargs,
    ):
        if params is None:
            params = PointMassParams()
        if device is None:
            device = torch.device("cpu")
        self._batch_size = num_envs
        self._params = params

        self._device = device
        self._goal = torch.tensor([0.0, 0.0]).to(self._device)
        self._ep_step = 0
        self._prev_dist_idx = -1

        self._ep_rewards = []
        obs_space = spaces.Box(low=-1.0, high=1.0, shape=(2,), seed=seed)
        ac_space = spaces.Box(low=-1.0, high=1.0, shape=(2,), seed=seed)

        self._is_eval = set_eval or self._params.force_eval_start_dist
        if self._params.force_train_start_dist:
            self._is_eval = False

        if self._is_eval:
            regions = self.get_regions(
                self._params.eval_offset, self._params.start_state_noise
            )
        else:
            regions = self.get_regions(
                self._params.train_offset, self._params.start_state_noise
            )

        if self._params.start_state_noise != 0:
            self._start_distributions = Uniform(regions[:, 0], regions[:, 1])
        else:
            self._start_distributions = SingleSampler(regions[:, 0])

        super().__init__(
            self._batch_size,
            obs_space,
            ac_space,
        )

    def step_async(self, actions):
        pass

    def step_wait(self):
        pass

    def forward(self, cur_pos, action):
        action = action.to(self._device)
        if self._params.clip_actions:
            action = torch.clamp(action, -1.0, 1.0)
        new_pos = cur_pos + (action * self._params.dt)
        if self._params.transition_noise != 0.0:
            action += self._params.transition_noise * torch.randn(
                action.shape, device=self._device
            )

        if self._params.clip_bounds:
            new_pos = torch.clamp(
                new_pos,
                -self._params.position_limit,
                self._params.position_limit,
            )
        return new_pos

    def step(self, action):
        self.cur_pos = self.forward(self.cur_pos, action)
        self._ep_step += 1
        self._store_actions.append(action)

        is_done = self._ep_step >= self._params.ep_horizon
        reward = self._get_reward(action)
        self._ep_rewards.append(reward)

        all_is_done = torch.tensor(
            [is_done for _ in range(self._batch_size)], dtype=torch.bool
        )
        dist_to_goal = torch.linalg.norm(
            self._goal - self.cur_pos, dim=-1, keepdims=True
        )

        all_info = [
            {"dist_to_goal": dist_to_goal[i].item()} for i in range(self._batch_size)
        ]
        all_info = self._add_to_info(all_info)

        if is_done:
            store_actions = torch.stack(self._store_actions, dim=1)
            action_magnitudes = torch.linalg.norm(store_actions, dim=-1)
            final_obs = self._get_obs()
            for i in range(self._batch_size):
                all_info[i]["episode"] = {
                    "r": torch.stack(self._ep_rewards).sum(0)[i].item(),
                    "max_action_magnitude": action_magnitudes[i].max().item(),
                    "avg_action_magnitude": action_magnitudes[i].mean().item(),
                }
                all_info[i][FINAL_OBS_KEY] = final_obs[i]
            self.reset()

        return (self._get_obs(), reward, all_is_done, all_info)

    def get_images(self, mode=None, img_dim=64, **kwargs) -> np.ndarray:
        def convert_coordinate(coord):
            # Normalize position to [0,1]
            norm_pos = (coord + self._params.position_limit) / (
                2 * self._params.position_limit
            )

            # Convert position to image space
            return (norm_pos * img_dim).to(torch.long)

        def write_to_img(img, pos, size, color):
            lower_x = max(pos[0] - size, 0)
            upper_x = min(pos[0] + size, img_dim)

            lower_y = max(pos[1] - size, 0)
            upper_y = min(pos[1] + size, img_dim)

            img[lower_x:upper_x, lower_y:upper_y] = color
            return img

        agent_pos = convert_coordinate(self.cur_pos)
        goal_pos = convert_coordinate(self._goal)
        entity_size = img_dim // 32

        img = np.full((self._batch_size, img_dim, img_dim, 3), 255, dtype=np.uint8)
        agent_color = [8, 143, 143]
        goal_color = [224, 17, 95]

        for env_i in range(self._batch_size):
            img[env_i] = write_to_img(
                img[env_i], agent_pos[env_i], entity_size, agent_color
            )
            img[env_i] = write_to_img(img[env_i], goal_pos, entity_size, goal_color)

        return img

    def _add_to_info(self, all_info):
        return all_info

    def _get_dist(self):
        return torch.linalg.norm(self._goal - self.cur_pos, dim=-1, keepdims=True)

    def _get_reward(self, action):
        if self._params.custom_reward is None:
            dist_to_goal = torch.linalg.norm(
                self._goal - self.cur_pos, dim=-1, keepdims=True
            )

            reward = -self._params.reward_dist_pen * dist_to_goal
        else:
            reward = self._params.custom_reward(self.cur_pos, self._prev_pos, action)
        self._prev_pos = self.cur_pos.detach().clone()
        return reward  # noqa: R504

    def get_regions(self, offset, spread):
        inc = np.pi / 2

        centers = [offset + i * inc for i in range(4)]

        return torch.tensor(
            [[center - spread, center + spread] for center in centers]
        ).to(self._device)

    def _get_dist_idx(self, batch_size):
        if not self._params.random_start_region_sample:
            new_dist_idx = self._prev_dist_idx + 1
            new_dist_idx = new_dist_idx % self._params.num_train_regions
            self._prev_dist_idx = new_dist_idx
            return torch.full((batch_size,), new_dist_idx)

        if self._is_eval:
            return torch.randint(0, 4, (batch_size,))
        else:
            if self._params.start_idx == -1:
                return torch.randint(0, self._params.num_train_regions, (batch_size,))
            else:
                return torch.tensor([self._params.start_idx]).repeat(batch_size)

    def _sample_start(self, batch_size, offset_start):
        idx = self._get_dist_idx(batch_size).to(self._device)
        samples = self._start_distributions.sample(idx.shape)
        ang = samples.gather(1, idx.view(-1, 1)).view(-1)

        return (
            torch.stack(
                [
                    self._params.radius * torch.cos(ang),
                    self._params.radius * torch.sin(ang),
                ],
                dim=-1,
            ).to(self._device)
            + offset_start
        )

    def reset(self):
        self.cur_pos = self._sample_start(self._batch_size, self._goal)
        self._ep_step = 0
        self._ep_rewards = []
        self._prev_pos = self.cur_pos.detach().clone()
        self._store_actions = []

        return self._get_obs()

    def _get_obs(self):
        return self.cur_pos.clone()


class SingleSampler:
    def __init__(self, point):
        self.point = point

    def sample(self, shape):
        return self.point.unsqueeze(0).repeat(shape[0], 1)



from .auto_line import line_plot, plot_legend
from .auto_table import plot_table
from .utils import combine_dicts_to_df, fig_save
from .wb_query import batch_query, query, query_s

__all__ = [
    "plot_table",
    "query",
    "query_s",
    "batch_query",
    "line_plot",
    "plot_legend",
    "fig_save",
    "combine_dicts_to_df",
]



"""
Helpers for dealing with vectorized environments.
"""

import os
import os.path as osp
import pickle
import random
import time
from collections import OrderedDict, defaultdict
from typing import Any, Callable, Dict, List, Tuple

import gym
import gym.spaces as spaces
import numpy as np
import torch


def set_seed(seed: int) -> None:
    """
    Sets the seed for numpy, python random, and pytorch.
    """
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def dict_sel(d: Dict[Any, List[Any]], idx: int) -> Dict[Any, Any]:
    ret_dict = {}
    for k in d:
        ret_dict[k] = d[k][idx]
    return ret_dict


def apply_to_dict(d: Dict[Any, List[Any]], fn: Callable[[Any], Any]) -> Dict[Any, Any]:
    ret_dict = {}
    for k in d:
        ret_dict[k] = fn(d[k])
    return ret_dict


def transpose_list_dict(arr: List[Dict]) -> Dict[Any, Any]:
    keys = arr[0].keys()
    ret_d = {k: [] for k in keys}
    for arr_ele in arr:
        for k in keys:
            ret_d[k].append(arr_ele[k])

    for k in keys:
        if isinstance(ret_d[k][0], torch.Tensor):
            ret_d[k] = torch.stack(ret_d[k])

    return ret_d


def transpose_dict_list(d: Dict[Any, Any]) -> List[Dict]:
    keys = list(d.keys())
    lens = [len(d[k]) for k in d]
    if len(set(lens)) != 1:
        raise ValueError("All lists must have equal sizes")

    # Assumes that all the lists are equal length.
    ret = []
    for i in range(lens[0]):
        ret.append({k: d[k][i] for k in keys})
    return ret


def group_trajectories(
    dones: torch.Tensor, **other_data: Dict[str, torch.Tensor]
) -> List[Dict[str, torch.Tensor]]:
    """
    Takes flat lists of data and groups them according to when dones=0. Used to
    go from a list of transitions to a list of trajectories.

    :param dones: An (N, 1) tensor where "False" or "0" value indicates when to
        stop the group.
    """
    for k, v in other_data.items():
        if v.size(0) != dones.size(0):
            raise ValueError(
                f"Key {k} is improper shape {v.shape} for dones of shape {dones.shape}"
            )

    trajs = []
    cur_traj = defaultdict(list)
    for i, done in enumerate(dones):
        for k, v in other_data.items():
            cur_traj[k].append(v[i])
        if done:
            trajs.append({k: torch.stack(v, dim=0) for k, v in cur_traj.items()})
            cur_traj = defaultdict(list)
    return trajs


def compress_dict(d: Dict[str, Any], pre="", sep=".") -> Dict[str, Any]:
    """
    Compresses a dictionary to only have one "level"
    """
    ret_d = {}
    for k, v in d.items():
        if isinstance(v, dict):
            ret_d.update(compress_dict(v, f"{pre}{k}{sep}"))
        else:
            ret_d[pre + k] = str(v)
    return ret_d


def compress_and_filter_dict(d: Dict[str, Any], pre="") -> Dict[str, Any]:
    """
    Compresses a dictionary to only have one "level"
    """
    ret_d = {}
    for k, v in d.items():
        if isinstance(v, dict):
            ret_d.update(compress_and_filter_dict(v, f"{k}."))
        elif isinstance(v, (float, int, np.float32, np.float64, np.uint, np.int32)):
            ret_d[pre + k] = v
    return ret_d


def copy_obs_dict(obs):
    """
    Deep-copy an observation dict.
    """
    return {k: np.copy(v) for k, v in obs.items()}


def dict_to_obs(obs_dict):
    """
    Convert an observation dict into a raw array if the
    original observation space was not a Dict space.
    """
    if set(obs_dict.keys()) == {None}:
        return obs_dict[None]
    return obs_dict


def obs_space_info(obs_space):
    """
    Get dict-structured information about a gym.Space.

    Returns:
      A tuple (keys, shapes, dtypes):
        keys: a list of dict keys.
        shapes: a dict mapping keys to shapes.
        dtypes: a dict mapping keys to dtypes.
    """
    if isinstance(obs_space, gym.spaces.Dict):
        assert isinstance(obs_space.spaces, OrderedDict)
        subspaces = obs_space.spaces
    else:
        subspaces = {None: obs_space}
    keys = []
    shapes = {}
    dtypes = {}
    for key, box in subspaces.items():
        keys.append(key)
        shapes[key] = box.shape
        dtypes[key] = box.dtype
    return keys, shapes, dtypes


def obs_to_dict(obs):
    """
    Convert an observation into a dict.
    """
    if isinstance(obs, dict):
        return obs
    return {None: obs}


def reshape_obs_space(obs_space: spaces.Box, new_shape: Tuple[int]) -> spaces.Box:
    assert isinstance(obs_space, spaces.Box)
    return spaces.Box(
        shape=new_shape,
        high=obs_space.low.reshape(-1)[0],
        low=obs_space.high.reshape(-1)[0],
        dtype=obs_space.dtype,
    )


def is_dict_obs(ob_space):
    return isinstance(ob_space, gym.spaces.Dict)


def update_obs_space(cur_space, update_obs_space):
    if is_dict_obs(cur_space):
        new_obs_space = cur_space.spaces
        new_obs_space["observation"] = update_obs_space
        return gym.spaces.Dict(new_obs_space)
    else:
        return update_obs_space


class StackHelper:
    """
    A helper for stacking observations.
    """

    def __init__(self, ob_shape, n_stack, device, n_procs=None):
        self.input_dim = ob_shape[0]
        self.n_procs = n_procs
        self.real_shape = (n_stack * self.input_dim, *ob_shape[1:])
        if self.n_procs is not None:
            self.stacked_obs = torch.zeros((n_procs, *self.real_shape))
            if device is not None:
                self.stacked_obs = self.stacked_obs.to(device)
        else:
            self.stacked_obs = np.zeros(self.real_shape)

    def update_obs(self, obs, dones=None, infos=None):
        """
        - obs: torch.tensor
        """
        if self.n_procs is not None:
            self.stacked_obs[:, : -self.input_dim] = self.stacked_obs[
                :, self.input_dim :
            ].clone()
            for (i, new) in enumerate(dones):
                if new:
                    self.stacked_obs[i] = 0
            self.stacked_obs[:, -self.input_dim :] = obs

            # Update info so the final observation frame stack has the final
            # observation as the final frame in the stack.
            for i in range(len(infos)):
                if "final_obs" in infos[i]:
                    new_final = torch.zeros(*self.stacked_obs.shape[1:])
                    new_final[:-1] = self.stacked_obs[i][1:]
                    new_final[-1] = torch.tensor(infos[i]["final_obs"]).to(
                        self.stacked_obs.device
                    )
                    infos[i]["final_obs"] = new_final
            return self.stacked_obs.clone(), infos
        else:
            self.stacked_obs[: -self.input_dim] = self.stacked_obs[
                self.input_dim :
            ].copy()
            self.stacked_obs[-self.input_dim :] = obs

            return self.stacked_obs.copy(), infos

    def reset(self, obs):
        if self.n_procs is not None:
            if torch.backends.cudnn.deterministic:
                self.stacked_obs = torch.zeros(self.stacked_obs.shape)
            else:
                self.stacked_obs.zero_()
            self.stacked_obs[:, -self.input_dim :] = obs
            return self.stacked_obs.clone()
        else:
            self.stacked_obs = np.zeros(self.stacked_obs.shape)
            self.stacked_obs[-self.input_dim :] = obs
            return self.stacked_obs.copy()

    def get_shape(self):
        return self.real_shape


def get_size_for_space(space: spaces.Space) -> int:
    """
    Returns the number of dimensions to represent a Gym space. If the space is discrete, this requires only 1 dimension.
    """
    if isinstance(space, spaces.Discrete):
        return 1
    elif isinstance(space, spaces.Box):
        return space.shape[0]
    else:
        raise ValueError(f"Space {space} not supported")


class CacheHelper:
    CACHE_PATH = "./data/cache"

    def __init__(self, cache_name, def_val=None, verbose=False, rel_dir=""):
        self.use_cache_path = osp.join(CacheHelper.CACHE_PATH, rel_dir)
        os.makedirs(self.use_cache_path, exist_ok=True)
        self.cache_id = osp.join(self.use_cache_path, f"{cache_name}.pickle")
        self.def_val = def_val
        self.verbose = verbose

    def exists(self):
        return osp.exists(self.cache_id)

    def load(self, load_depth=0):
        if self.exists():
            try:
                with open(self.cache_id, "rb") as f:
                    if self.verbose:
                        print("Loading cache @", self.cache_id)
                    return pickle.load(f)
            except EOFError as e:
                if load_depth == 32:
                    raise e
                # try again soon
                print(
                    "Cache size is ",
                    osp.getsize(self.cache_id),
                    "for ",
                    self.cache_id,
                )
                time.sleep(1.0 + np.random.uniform(0.0, 1.0))
                return self.load(load_depth + 1)
            return self.def_val
        else:
            return self.def_val

    def save(self, val):
        with open(self.cache_id, "wb") as f:
            if self.verbose:
                print("Saving cache @", self.cache_id)
            pickle.dump(val, f)



import torch


def tensor_hash(X: torch.Tensor) -> int:
    """
    Returns a unique representation of a tensor. Warning, this will be slow for
    large tensors.
    """
    flat_X = X.detach().view(-1).cpu().tolist()
    return hash(tuple(flat_X))



from .dummy_vec_env import DummyVecEnv
from .shmem_vec_env import ShmemVecEnv
from .subproc_vec_env import SubprocVecEnv
from .vec_env import (
    AlreadySteppingError,
    CloudpickleWrapper,
    NotSteppingError,
    VecEnv,
    VecEnvObservationWrapper,
    VecEnvWrapper,
)
from .vec_frame_stack import VecFrameStack
from .vec_monitor import VecMonitor

__all__ = [
    "AlreadySteppingError",
    "NotSteppingError",
    "VecEnv",
    "VecEnvWrapper",
    "VecEnvObservationWrapper",
    "CloudpickleWrapper",
    "DummyVecEnv",
    "ShmemVecEnv",
    "SubprocVecEnv",
    "VecFrameStack",
    "VecMonitor",
]



import numpy as np


def tile_images(img_nhwc):
    """
    Tile N images into one big PxQ image
    (P,Q) are chosen to be as close as possible, and if N
    is square, then P=Q.

    input: img_nhwc, list or array of images, ndim=4 once turned into array
        n = batch index, h = height, w = width, c = channel
    returns:
        bigim_HWc, ndarray with ndim=3
    """
    img_nhwc = np.asarray(img_nhwc)

    # NOTE: Added to support renderings that return images series.
    if len(img_nhwc.shape) == 1:
        # Pad other sequences so they are as long.
        dims = [img_nhwc[i].shape[0] for i in range(img_nhwc.shape[0])]
        max_dim = max(dims)
        for i in range(img_nhwc.shape[0]):
            if img_nhwc[i].shape[0] < max_dim:
                img_nhwc[i] = np.pad(
                    img_nhwc[i], ((0, 1), (0, 0), (0, 0), (0, 0)), "edge"
                )
        img_nhwc = np.array(list(img_nhwc))
        return tile_images(img_nhwc)

    if len(img_nhwc.shape) == 5:
        return np.array([tile_images(img_nhwc[:, i]) for i in range(img_nhwc.shape[1])])

    N, h, w, c = img_nhwc.shape

    H = int(np.ceil(np.sqrt(N)))
    W = int(np.ceil(float(N) / H))
    img_nhwc = np.array(list(img_nhwc) + [img_nhwc[0] * 0 for _ in range(N, H * W)])
    img_HWhwc = img_nhwc.reshape(H, W, h, w, c)
    # add boundary to distinguish environments
    img_HWhwc[:, :, :, -1, :] = 0
    img_HWhwc[:, :, -1, :, :] = 0
    img_HhWwc = img_HWhwc.transpose(0, 2, 1, 3, 4)
    return img_HhWwc.reshape(H * h, W * w, c)



import os
import os.path as osp
from typing import Any, Dict, List

import matplotlib.pyplot as plt
import pandas as pd

try:
    import wandb
except ImportError:
    wandb = None

MISSING_VALUE = 0.2444


def combine_dicts_to_df(
    data: List[Dict[str, Any]], existing_df_key: str
) -> pd.DataFrame:
    all_dfs = []
    for r in data:
        use_df = r[existing_df_key]
        for k, v in r.items():
            if k == existing_df_key:
                continue
            use_df[k] = v
        all_dfs.append(use_df)
    return pd.concat(all_dfs)


def fig_save(
    save_dir,
    save_name,
    fig,
    is_high_quality=True,
    verbose=True,
    clear=False,
    log_wandb=False,
    wandb_name=None,
) -> str:
    """
    :param save_dir: Directory to save file to. Directory is created if it does not exist.
    :param save_name: No file extension included in name.
    :returns: The saved full file path.
    """
    if not osp.exists(save_dir):
        os.makedirs(save_dir)

    plt.tight_layout()
    if is_high_quality:
        full_path = osp.join(save_dir, f"{save_name}.pdf")
        fig.savefig(full_path, bbox_inches="tight", dpi=100)
    else:
        full_path = osp.join(save_dir, f"{save_name}.png")
        fig.savefig(full_path)

    if verbose:
        print(f"Saved to {full_path}")
    if clear:
        plt.close(fig)
        plt.clf()

    if log_wandb:
        if wandb_name is None:
            raise ValueError("Must specify wandb log name as well")
        if wandb is None:
            raise ValueError("Wandb is not installed.")
        wandb.log({wandb_name: wandb.Image(full_path)})
    return full_path



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Code is from https://github.com/facebookresearch/habitat-lab/blob/main/habitat_baselines/rl/models/rnn_state_encoder.py
"""

from typing import Tuple

import torch
import torch.nn as nn
from torch.nn.utils.rnn import PackedSequence


def _invert_permutation(permutation: torch.Tensor) -> torch.Tensor:
    output = torch.empty_like(permutation)
    output.scatter_(
        0,
        permutation,
        torch.arange(0, permutation.numel(), device=permutation.device),
    )
    return output


def _build_pack_info_from_dones(
    dones: torch.Tensor,
    T: int,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""Create the indexing info needed to make the PackedSequence
    based on the dones.

    PackedSequences are PyTorch's way of supporting a single RNN forward
    call where each input in the batch can have an arbitrary sequence length

    They work as follows: Given the sequences [c], [x, y, z], [a, b],
    we generate data [x, a, c, y, b, z] and batch_sizes [3, 2, 1].  The
    data is a flattened out version of the input sequences (the ordering in
    data is determined by sequence length).  batch_sizes tells you that
    for each index, how many sequences have a length of (index + 1) or greater.

    This method will generate the new index ordering such that you can
    construct the data for a PackedSequence from a (T*N, ...) tensor
    via x.index_select(0, select_inds)
    """
    dones = dones.view(T, -1)
    N = dones.size(1)

    rollout_boundaries = dones.clone().detach()
    # Force a rollout boundary for t=0.  We will use the
    # original dones for masking later, so this is fine
    # and simplifies logic considerably
    rollout_boundaries[0] = True
    rollout_boundaries = rollout_boundaries.nonzero(as_tuple=False)

    # The rollout_boundaries[:, 0]*N will make the episode_starts index into
    # the T*N flattened tensors
    episode_starts = rollout_boundaries[:, 0] * N + rollout_boundaries[:, 1]

    # We need to create a transposed start indexing so we can compute episode lengths
    # As if we make the starts index into a N*T tensor, then starts[1] - starts[0]
    # will compute the length of the 0th episode
    episode_starts_transposed = rollout_boundaries[:, 1] * T + rollout_boundaries[:, 0]
    # Need to sort so the above logic is correct
    episode_starts_transposed, sorted_indices = torch.sort(
        episode_starts_transposed, descending=False
    )

    # Calculate length of episode rollouts
    rollout_lengths = episode_starts_transposed[1:] - episode_starts_transposed[:-1]
    last_len = N * T - episode_starts_transposed[-1]
    rollout_lengths = torch.cat([rollout_lengths, last_len.unsqueeze(0)])
    # Undo the sort above
    rollout_lengths = rollout_lengths.index_select(
        0, _invert_permutation(sorted_indices)
    )

    # Resort in descending order of episode length
    lengths, sorted_indices = torch.sort(rollout_lengths, descending=True)

    # We will want these on the CPU for torch.unique_consecutive,
    # so move now.
    cpu_lengths = lengths.to(device="cpu", non_blocking=True)

    episode_starts = episode_starts.index_select(0, sorted_indices)
    select_inds = torch.empty((T * N), device=dones.device, dtype=torch.int64)

    max_length = int(cpu_lengths[0].item())
    # batch_sizes is *always* on the CPU
    batch_sizes = torch.empty((max_length,), device="cpu", dtype=torch.long)

    offset = 0
    prev_len = 0
    num_valid_for_length = lengths.size(0)

    unique_lengths = torch.unique_consecutive(cpu_lengths)
    # Iterate over all unique lengths in reverse as they sorted
    # in decreasing order
    for next_len in reversed(unique_lengths):
        valids = lengths[0:num_valid_for_length] > prev_len
        num_valid_for_length = int(valids.float().sum())

        batch_sizes[prev_len:next_len] = num_valid_for_length

        # Creates this array
        # [step * N + start for step in range(prev_len, next_len)
        #                   for start in episode_starts[0:num_valid_for_length]
        # * N because each step is separated by N elements
        new_inds = (
            torch.arange(prev_len, next_len, device=episode_starts.device).view(
                next_len - prev_len, 1
            )
            * N
            + episode_starts[0:num_valid_for_length].view(1, num_valid_for_length)
        ).view(-1)

        select_inds[offset : offset + new_inds.numel()] = new_inds

        offset += new_inds.numel()

        prev_len = next_len

    # Make sure we have an index for all elements
    assert offset == T * N

    # This is used in conjunction with episode_starts to get
    # the RNN hidden states
    rnn_state_batch_inds = episode_starts % N
    # This indicates that a given episode is the last one
    # in that rollout.  In other words, there are N places
    # where this is True, and for each n, True indicates
    # that this episode is the last contiguous block of experience,
    # This is needed for getting the correct hidden states after
    # the RNN forward pass
    last_episode_in_batch_mask = ((episode_starts + (lengths - 1) * N) // N) == (T - 1)

    return (
        select_inds,
        batch_sizes,
        episode_starts,
        rnn_state_batch_inds,
        last_episode_in_batch_mask,
    )


def build_rnn_inputs(
    x: torch.Tensor, not_dones: torch.Tensor, rnn_states: torch.Tensor
) -> Tuple[PackedSequence, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""Create a PackedSequence input for an RNN such that each
    set of steps that are part of the same episode are all part of
    a batch in the PackedSequence.

    Use the returned select_inds and build_rnn_out_from_seq to invert this.

    :param x: A (T * N, -1) tensor of the data to build the PackedSequence out of
    :param not_dones: A (T * N) tensor where not_dones[i] == False indicates an episode is done
    :param rnn_states: A (-1, N, -1) tensor of the rnn_hidden_states

    :return: tuple(x_seq, rnn_states, select_inds, rnn_state_batch_inds, last_episode_in_batch_mask)
        WHERE
        x_seq is the PackedSequence version of x to pass to the RNN

        rnn_states are the corresponding rnn state

        select_inds can be passed to build_rnn_out_from_seq to retrieve the
            RNN output

        rnn_state_batch_inds indicates which of the rollouts in the batch a hidden
            state came from/is for

        last_episode_in_batch_mask indicates if an episode is the last in that batch.
            There will be exactly N places where this is True

    """

    N = rnn_states.size(1)
    T = x.size(0) // N
    dones = torch.logical_not(not_dones)

    (
        select_inds,
        batch_sizes,
        episode_starts,
        rnn_state_batch_inds,
        last_episode_in_batch_mask,
    ) = _build_pack_info_from_dones(dones.detach().to(device="cpu"), T)

    select_inds = select_inds.to(device=x.device)
    episode_starts = episode_starts.to(device=x.device)
    rnn_state_batch_inds = rnn_state_batch_inds.to(device=x.device)
    last_episode_in_batch_mask = last_episode_in_batch_mask.to(device=x.device)

    x_seq = PackedSequence(x.index_select(0, select_inds), batch_sizes, None, None)

    # Just select the rnn_states by batch index, the masking bellow will set things
    # to zero in the correct locations
    rnn_states = rnn_states.index_select(1, rnn_state_batch_inds)
    # Now zero things out in the correct locations
    rnn_states = torch.where(
        not_dones.view(1, -1, 1).index_select(1, episode_starts),
        rnn_states,
        rnn_states.new_zeros(()),
    )

    return (
        x_seq,
        rnn_states,
        select_inds,
        rnn_state_batch_inds,
        last_episode_in_batch_mask,
    )


def build_rnn_out_from_seq(
    x_seq: PackedSequence,
    hidden_states,
    select_inds,
    rnn_state_batch_inds,
    last_episode_in_batch_mask,
    N: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""Construct the output of the rnn from a packed sequence returned by
        forward propping an RNN on the packed sequence returned by :ref:`build_rnn_inputs`.

    :param x_seq: The packed sequence output from the rnn
    :param hidden_statess: The hidden states output from the rnn
    :param select_inds: Returned from :ref:`build_rnn_inputs`
    :param rnn_state_batch_inds: Returned from :ref:`build_rnn_inputs`
    :param last_episode_in_batch_mask: Returned from :ref:`build_rnn_inputs`
    :param N: The number of simulator instances in the batch of experience.
    """
    x = x_seq.data.index_select(0, _invert_permutation(select_inds))

    last_hidden_states = torch.masked_select(
        hidden_states,
        last_episode_in_batch_mask.view(1, hidden_states.size(1), 1),
    ).view(hidden_states.size(0), N, hidden_states.size(2))
    output_hidden_states = torch.empty_like(last_hidden_states)
    scatter_inds = (
        torch.masked_select(rnn_state_batch_inds, last_episode_in_batch_mask)
        .view(1, N, 1)
        .expand_as(output_hidden_states)
    )
    output_hidden_states.scatter_(1, scatter_inds, last_hidden_states)

    return x, output_hidden_states


class RNNStateEncoder(nn.Module):
    r"""RNN encoder for use with RL and possibly IL.

    The main functionality this provides over just using PyTorch's RNN interface directly
    is that it takes an addition masks input that resets the hidden state between two adjacent
    timesteps to handle episodes ending in the middle of a rollout.
    """
    rnn: torch.nn.Module

    def layer_init(self):
        for name, param in self.rnn.named_parameters():
            if "weight" in name:
                nn.init.orthogonal_(param)
            elif "bias" in name:
                nn.init.constant_(param, 0)

    def pack_hidden(self, hidden_states: torch.Tensor) -> torch.Tensor:
        return hidden_states

    def unpack_hidden(self, hidden_states: torch.Tensor) -> torch.Tensor:
        return hidden_states

    def single_forward(
        self, x, hidden_states, masks
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Forward for a non-sequence input"""

        hidden_states = torch.where(
            masks.view(1, -1, 1), hidden_states, hidden_states.new_zeros(())
        )

        x, hidden_states = self.rnn(x.unsqueeze(0), self.unpack_hidden(hidden_states))
        hidden_states = self.pack_hidden(hidden_states)

        x = x.squeeze(0)
        return x, hidden_states

    def seq_forward(self, x, hidden_states, masks) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Forward for a sequence of length T

        Args:
            x: (T, N, -1) Tensor that has been flattened to (T * N, -1)
            hidden_states: The starting hidden state.
            masks: The masks to be applied to hidden state at every timestep.
                A (T, N) tensor flatten to (T * N)
        """
        N = hidden_states.size(1)

        (
            x_seq,
            hidden_states,
            select_inds,
            rnn_state_batch_inds,
            last_episode_in_batch_mask,
        ) = build_rnn_inputs(x, masks, hidden_states)

        x_seq, hidden_states = self.rnn(x_seq, self.unpack_hidden(hidden_states))
        hidden_states = self.pack_hidden(hidden_states)

        x, hidden_states = build_rnn_out_from_seq(
            x_seq,
            hidden_states,
            select_inds,
            rnn_state_batch_inds,
            last_episode_in_batch_mask,
            N,
        )

        return x, hidden_states

    def forward(self, x, hidden_states, masks) -> Tuple[torch.Tensor, torch.Tensor]:
        hidden_states = hidden_states.permute(1, 0, 2)
        if x.size(0) == hidden_states.size(1):
            x, hidden_states = self.single_forward(x, hidden_states, masks)
        else:
            x, hidden_states = self.seq_forward(x, hidden_states, masks)

        hidden_states = hidden_states.permute(1, 0, 2)

        return x, hidden_states


class LSTMStateEncoder(RNNStateEncoder):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
    ):
        super().__init__()

        self.num_recurrent_layers = num_layers * 2

        self.rnn = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
        )

        self.layer_init()

    def pack_hidden(
        self, hidden_states: Tuple[torch.Tensor, torch.Tensor]
    ) -> torch.Tensor:
        return torch.cat(hidden_states, 0)

    def unpack_hidden(self, hidden_states) -> Tuple[torch.Tensor, torch.Tensor]:
        lstm_states = torch.chunk(hidden_states, 2, 0)
        return (lstm_states[0], lstm_states[1])


class GRUStateEncoder(RNNStateEncoder):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
    ):
        super().__init__()

        self.num_recurrent_layers = num_layers

        self.rnn = nn.GRU(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
        )

        self.layer_init()


def build_rnn_state_encoder(
    input_size: int,
    hidden_size: int,
    rnn_type: str = "GRU",
    num_layers: int = 1,
):
    r"""Factory for :ref:`RNNStateEncoder`.  Returns one with either a GRU or LSTM based on
        the specified RNN type.

    :param input_size: The input size of the RNN
    :param hidden_size: The hidden dimension of the RNN
    :param rnn_types: The type of the RNN cell.  Can either be GRU or LSTM
    :param num_layers: The number of RNN layers.
    """
    rnn_type = rnn_type.lower()
    if rnn_type == "gru":
        return GRUStateEncoder(input_size, hidden_size, num_layers)
    elif rnn_type == "lstm":
        return LSTMStateEncoder(input_size, hidden_size, num_layers)
    else:
        raise RuntimeError(f"Did not recognize rnn type '{rnn_type}'")



from dataclasses import dataclass, field
from typing import List, Optional, Tuple

import numpy as np
import torch

from rl_utils.envs.pointmass.pointmass_env import PointMassEnv, PointMassParams
from rl_utils.envs.registry import full_env_registry


@dataclass(frozen=True)
class SquareObstacle:
    """
    * x,y position of the CENTER of the square.
    * Obstacle width
    * Obstacle length
    * Obstacle rotation angle (in degrees)
    """

    xy: Tuple[float, float]
    width: float
    height: float
    rot_deg: float


@dataclass(frozen=True)
class PointMassObstacleParams(PointMassParams):
    goal_thresh: float = 0.05
    square_obstacles: List[SquareObstacle] = field(default_factory=list)


@full_env_registry.register_env("PointMassObstacle-v0")
class PointMassObstacleEnv(PointMassEnv):
    def __init__(
        self,
        num_envs: int,
        params: Optional[PointMassObstacleParams] = None,
        device: Optional[torch.device] = None,
        set_eval: bool = False,
        seed: Optional[int] = None,
        **kwargs,
    ):
        if params is None:
            params = PointMassObstacleParams()
        super().__init__(num_envs, params, device, set_eval, seed, **kwargs)
        self._circle_obs = []
        self._square_obs_T = []

        for obs in self._params.square_obstacles:
            rot = obs.rot_deg * (np.pi / 180.0)

            rot_T = torch.tensor(
                [
                    [np.cos(rot), -np.sin(rot), 0.0],
                    [np.sin(rot), np.cos(rot), 0.0],
                    [0.0, 0.0, 1.0],
                ],
                device=self._device,
                dtype=torch.float,
            )
            trans_T = torch.tensor(
                [
                    [1.0, 0.0, obs.xy[0]],
                    [0.0, 1.0, obs.xy[1]],
                    [0.0, 0.0, 1.0],
                ],
                device=self._device,
                dtype=torch.float,
            )

            self._square_obs_T.append(
                (
                    trans_T @ rot_T,
                    obs.width,
                    obs.height,
                )
            )

    def _add_to_info(self, all_info):
        dists = self._get_dist()
        for i in range(self._batch_size):
            all_info[i]["at_goal"] = dists[i].item() < self._params.goal_thresh
        return all_info

    def forward(self, cur_pos, action):
        action = action.to(self._device)
        action = torch.clamp(action, -1.0, 1.0)
        new_pos = cur_pos + (action * self._params.dt)

        if self._params.clip_bounds:
            new_pos = torch.clamp(
                new_pos,
                -self._params.position_limit,
                self._params.position_limit,
            )

        for ob_pos, ob_radius in self._circle_obs:
            local_pos = new_pos - ob_pos
            local_dist = torch.linalg.norm(local_pos, dim=-1)
            coll_idxs = torch.nonzero(local_dist < ob_radius)

            norm_pos = (local_pos / local_dist.view(-1, 1)) * ob_radius
            adjusted_pos = ob_pos + norm_pos
            new_pos[coll_idxs] = adjusted_pos[coll_idxs]

        inside_obstacle = self.is_inside_obstacle(new_pos)

        new_pos[inside_obstacle] = cur_pos[inside_obstacle]

        return new_pos

    def is_inside_obstacle(self, pos: torch.Tensor) -> torch.BoolTensor:
        """
        :param pos: A tensor of shape (N, 2).
        :returns: Tensor of shape (N,) indicating if the points were inside the obstacle.
        """

        homo_pos = torch.cat(
            [pos, torch.ones(pos.shape[0], 1, device=self._device)], dim=-1
        )
        inside_any_box = torch.zeros(
            pos.shape[0], device=self._device, dtype=torch.bool
        )
        for obs_T, xlen, ylen in self._square_obs_T:
            local_pos = torch.linalg.inv(obs_T) @ homo_pos.T

            inside_x = torch.logical_and(
                local_pos[0] < (xlen / 2), local_pos[0] > -(xlen / 2)
            )
            inside_y = torch.logical_and(
                local_pos[1] < (ylen / 2), local_pos[1] > -(ylen / 2)
            )
            inside_box = torch.logical_and(inside_x, inside_y)
            inside_any_box |= inside_box
        return inside_any_box



"""
An interface for asynchronous vectorized environments.
"""

import ctypes
import multiprocessing as mp
from collections.abc import Iterable

import numpy as np

from rl_utils.common.core_utils import dict_to_obs, obs_space_info, obs_to_dict

from .vec_env import FINAL_OBS_KEY, CloudpickleWrapper, VecEnv, clear_mpi_env_vars

_NP_TO_CT = {
    np.float32: ctypes.c_float,
    np.float64: ctypes.c_double,
    np.int32: ctypes.c_int32,
    np.int8: ctypes.c_int8,
    np.uint8: ctypes.c_char,
    bool: ctypes.c_bool,
}


class ShmemVecEnv(VecEnv):
    """
    Optimized version of SubprocVecEnv that uses shared variables to communicate observations.
    """

    def __init__(self, env_fns, spaces=None, context="spawn"):
        """
        If you don't specify observation_space, we'll have to create a dummy
        environment to get it.
        """
        ctx = mp.get_context(context)
        if spaces:
            observation_space, action_space = spaces
        else:
            dummy = env_fns[0]()
            observation_space, action_space = (
                dummy.observation_space,
                dummy.action_space,
            )
            dummy.close()
            del dummy
        VecEnv.__init__(self, len(env_fns), observation_space, action_space)
        self.obs_keys, self.obs_shapes, self.obs_dtypes = obs_space_info(
            observation_space
        )
        self.obs_bufs = [
            {
                k: ctx.Array(
                    _NP_TO_CT[self.obs_dtypes[k].type],
                    int(np.prod(self.obs_shapes[k])),
                )
                for k in self.obs_keys
            }
            for _ in env_fns
        ]
        self.parent_pipes = []
        self.procs = []
        with clear_mpi_env_vars():
            for env_fn, obs_buf in zip(env_fns, self.obs_bufs):
                wrapped_fn = CloudpickleWrapper(env_fn)
                parent_pipe, child_pipe = ctx.Pipe()
                proc = ctx.Process(
                    target=_subproc_worker,
                    args=(
                        child_pipe,
                        parent_pipe,
                        wrapped_fn,
                        obs_buf,
                        self.obs_shapes,
                        self.obs_dtypes,
                        self.obs_keys,
                    ),
                )
                proc.daemon = True
                self.procs.append(proc)
                self.parent_pipes.append(parent_pipe)
                proc.start()
                child_pipe.close()
        self.waiting_step = False
        self.viewer = None

    def reset(self):
        if self.waiting_step:
            print("Called reset() while waiting for the step to complete")
            self.step_wait()
        for pipe in self.parent_pipes:
            pipe.send(("reset", None))
        return self._decode_obses([pipe.recv() for pipe in self.parent_pipes])

    def step_async(self, actions):
        assert len(actions) == len(self.parent_pipes)
        for pipe, act in zip(self.parent_pipes, actions):
            pipe.send(("step", act))

    def step_wait(self):
        outs = [pipe.recv() for pipe in self.parent_pipes]
        obs, rews, dones, infos = zip(*outs)
        return self._decode_obses(obs), np.array(rews), np.array(dones), infos

    def close_extras(self):
        if self.waiting_step:
            self.step_wait()
        for pipe in self.parent_pipes:
            pipe.send(("close", None))
        for pipe in self.parent_pipes:
            pipe.recv()
            pipe.close()
        for proc in self.procs:
            proc.join()

    def get_images(self, mode="human", **kwargs):
        N = len(self.parent_pipes)
        all_pipe_kwargs = []
        for i in range(N):
            pipe_kwargs = {}
            for k, v in kwargs.items():
                if isinstance(v, Iterable):
                    pipe_kwargs[k] = kwargs[k][i]
                else:
                    pipe_kwargs[k] = kwargs[k]
            all_pipe_kwargs.append(pipe_kwargs)

        for pipe, p_kwargs in zip(self.parent_pipes, all_pipe_kwargs):
            pipe.send(("render", (mode, p_kwargs)))
        return [pipe.recv() for pipe in self.parent_pipes]

    def _decode_obses(self, obs):
        result = {}
        for k in self.obs_keys:
            bufs = [b[k] for b in self.obs_bufs]
            o = [
                np.frombuffer(b.get_obj(), dtype=self.obs_dtypes[k]).reshape(
                    self.obs_shapes[k]
                )
                for b in bufs
            ]
            result[k] = np.array(o)
        return dict_to_obs(result)


def _subproc_worker(
    pipe, parent_pipe, env_fn_wrapper, obs_bufs, obs_shapes, obs_dtypes, keys
):
    """
    Control a single environment instance using IPC and
    shared memory.
    """

    def _write_obs(maybe_dict_obs):
        flatdict = obs_to_dict(maybe_dict_obs)
        for k in keys:
            dst = obs_bufs[k].get_obj()
            dst_np = np.frombuffer(dst, dtype=obs_dtypes[k]).reshape(
                obs_shapes[k]
            )  # pylint: disable=W0212
            np.copyto(dst_np, flatdict[k])

    env = env_fn_wrapper.x()
    parent_pipe.close()
    try:
        while True:
            cmd, data = pipe.recv()
            if cmd == "reset":
                pipe.send(_write_obs(env.reset()))
            elif cmd == "step":
                obs, reward, done, info = env.step(data)
                if done:
                    final_obs = obs
                    info[FINAL_OBS_KEY] = final_obs
                    obs = env.reset()
                pipe.send((_write_obs(obs), reward, done, info))
            elif cmd == "render":
                pipe.send(env.render(mode=data[0], **data[1]))
            elif cmd == "close":
                pipe.send(None)
                break
            else:
                raise RuntimeError("Got unrecognized cmd %s" % cmd)
    except KeyboardInterrupt:
        print("ShmemVecEnv worker: got KeyboardInterrupt")
    finally:
        env.close()



import gym.spaces as spaces
import torch

import rl_utils.common as cutils

from .vec_env import VecEnvObservationWrapper, VecEnvWrapper


class VecEnvClipActions(VecEnvWrapper):
    def __init__(self, venv):
        obs_space = venv.action_space
        self._ac_low = torch.from_numpy(obs_space.low)
        self._ac_high = torch.from_numpy(obs_space.high)
        VecEnvWrapper.__init__(self, venv)

    def step(self, actions):
        actions = torch.clip(actions, self._ac_low, self._ac_high)
        return super().step(actions)


class VecEnvPermuteFrames(VecEnvObservationWrapper):
    def __init__(self, venv):
        obs_space = venv.observation_space
        if not isinstance(obs_space, spaces.Dict):
            obs_space = spaces.Dict({None: obs_space})

        permute_ks = {k: v for k, v in obs_space.spaces.items() if len(v.shape) == 3}
        for k, v in permute_ks.items():
            s = v.shape
            obs_space[k] = cutils.reshape_obs_space(v, (s[2], s[0], s[1]))

        VecEnvWrapper.__init__(self, venv, observation_space=obs_space)

    def process(self, obs):
        return obs.permute(0, 3, 1, 2)



import gym
import numpy as np
import torch

import rl_utils.common.core_utils as utils
from rl_utils.envs.vec_env.vec_env import VecEnvWrapper


# Checks whether done was caused my timit limits or not
class TimeLimitMask(gym.Wrapper):
    def step(self, action):
        obs, rew, done, info = self.env.step(action)
        if done and self.env._max_episode_steps == self.env._elapsed_steps:
            info["bad_transition"] = True

        return obs, rew, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


class VecPyTorch(VecEnvWrapper):
    def __init__(self, venv, device):
        super(VecPyTorch, self).__init__(venv)
        self.device = device

    def _data_convert(self, arr):
        if isinstance(arr, np.ndarray) and arr.dtype == np.float64:
            return arr.astype(np.float32)
        return arr

    def reset(self):
        obs = self.venv.reset()
        return self._trans_obs(obs)

    def step_async(self, actions):
        if isinstance(actions, torch.LongTensor):
            # Squeeze the dimension for discrete actions
            actions = actions.squeeze(1)
        actions = actions.cpu().numpy()
        self.venv.step_async(actions)

    def _trans_obs(self, obs):
        # Support for dict observations
        def _convert_obs(x):
            x = self._data_convert(x)
            x = torch.Tensor(x)
            return x.to(self.device)

        if isinstance(obs, dict):
            for k in obs:
                obs[k] = _convert_obs(obs[k])
        else:
            return _convert_obs(obs)
        return obs

    def step_wait(self):
        obs, reward, done, info = self.venv.step_wait()
        obs = self._trans_obs(obs)

        reward = torch.Tensor(reward).unsqueeze(dim=1)
        # Reward is sometimes a Double. Observation is considered to always be
        # float32
        reward = reward.float()
        return obs, reward, done, info


# Derived from
# https://github.com/openai/baselines/blob/master/baselines/common/vec_env/vec_frame_stack.py
class VecPyTorchFrameStack(VecEnvWrapper):
    """
    For now, this will only stack the "observation" key in dictionary
    observation spaces.
    """

    def __init__(self, venv, nstack, device):
        self.venv = venv
        self.nstack = nstack

        ob_space = venv.observation_space

        self.stacked_obs = utils.StackHelper(
            ob_space.shape, nstack, device, venv.num_envs
        )
        new_obs_space = utils.update_obs_space(
            venv.observation_space,
            utils.reshape_obs_space(ob_space, self.stacked_obs.get_shape()),
        )

        VecEnvWrapper.__init__(self, venv, observation_space=new_obs_space)

    def step_wait(self):
        obs, rews, news, infos = self.venv.step_wait()

        stacked_obs, infos = self.stacked_obs.update_obs(obs, news, infos)

        return stacked_obs, rews, news, infos

    def reset(self):
        obs = self.venv.reset()
        stacked_obs = self.stacked_obs.reset(obs)
        return utils.set_def_obs(obs, stacked_obs)

    def close(self):
        self.venv.close()



import multiprocessing as mp

import numpy as np

from .vec_env import CloudpickleWrapper, VecEnv, clear_mpi_env_vars


def worker(remote, parent_remote, env_fn_wrapper):
    parent_remote.close()
    env = env_fn_wrapper.x()
    try:
        while True:
            cmd, data = remote.recv()
            if cmd == "step":
                ob, reward, done, info = env.step(data)
                if done:
                    ob = env.reset()
                remote.send((ob, reward, done, info))
            elif cmd == "reset":
                ob = env.reset()
                remote.send(ob)
            elif cmd == "render":
                remote.send(env.render(mode="rgb_array"))
            elif cmd == "close":
                remote.close()
                break
            elif cmd == "get_spaces_spec":
                remote.send((env.observation_space, env.action_space, env.spec))
            else:
                raise NotImplementedError
    except KeyboardInterrupt:
        print("SubprocVecEnv worker: got KeyboardInterrupt")
    finally:
        env.close()


class SubprocVecEnv(VecEnv):
    """
    VecEnv that runs multiple environments in parallel in subproceses and communicates with them via pipes.
    Recommended to use when num_envs > 1 and step() can be a bottleneck.
    """

    def __init__(self, env_fns, spaces=None, context="spawn"):
        """
        Arguments:

        env_fns: iterable of callables -  functions that create environments to run in subprocesses. Need to be cloud-pickleable
        """
        self.waiting = False
        self.closed = False
        nenvs = len(env_fns)
        ctx = mp.get_context(context)
        self.remotes, self.work_remotes = zip(*[ctx.Pipe() for _ in range(nenvs)])
        self.ps = [
            ctx.Process(
                target=worker, args=(work_remote, remote, CloudpickleWrapper(env_fn))
            )
            for (work_remote, remote, env_fn) in zip(
                self.work_remotes, self.remotes, env_fns
            )
        ]
        for p in self.ps:
            p.daemon = (
                True  # if the main process crashes, we should not cause things to hang
            )
            with clear_mpi_env_vars():
                p.start()
        for remote in self.work_remotes:
            remote.close()

        self.remotes[0].send(("get_spaces_spec", None))
        observation_space, action_space, self.spec = self.remotes[0].recv()
        self.viewer = None
        VecEnv.__init__(self, len(env_fns), observation_space, action_space)

    def step_async(self, actions):
        self._assert_not_closed()
        for remote, action in zip(self.remotes, actions):
            remote.send(("step", action))
        self.waiting = True

    def step_wait(self):
        self._assert_not_closed()
        results = [remote.recv() for remote in self.remotes]
        self.waiting = False
        obs, rews, dones, infos = zip(*results)
        return _flatten_obs(obs), np.stack(rews), np.stack(dones), infos

    def reset(self):
        self._assert_not_closed()
        for remote in self.remotes:
            remote.send(("reset", None))
        return _flatten_obs([remote.recv() for remote in self.remotes])

    def close_extras(self):
        self.closed = True
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
        for remote in self.remotes:
            remote.send(("close", None))
        for p in self.ps:
            p.join()

    def get_images(self):
        self._assert_not_closed()
        for pipe in self.remotes:
            pipe.send(("render", None))
        return [pipe.recv() for pipe in self.remotes]

    def _assert_not_closed(self):
        assert (
            not self.closed
        ), "Trying to operate on a SubprocVecEnv after calling close()"

    def __del__(self):
        if not self.closed:
            self.close()


def _flatten_obs(obs):
    assert isinstance(obs, (list, tuple))
    assert len(obs) > 0

    if isinstance(obs[0], dict):
        keys = obs[0].keys()
        return {k: np.stack([o[k] for o in obs]) for k in keys}
    else:
        return np.stack(obs)



from .distributions import FixedCategorical, FixedNormal
from .rnn_state_encoder import (
    GRUStateEncoder,
    LSTMStateEncoder,
    RNNStateEncoder,
    build_rnn_state_encoder,
)
from .simple_cnn import SimpleCNN

__all__ = [
    "SimpleCNN",
    "GRUStateEncoder",
    "LSTMStateEncoder",
    "RNNStateEncoder",
    "build_rnn_state_encoder",
    "FixedCategorical",
    "FixedNormal",
]



from typing import Dict, List, Optional, Tuple

import torch
from torch.utils.data import Dataset


class DictDataset(Dataset):
    def __init__(
        self,
        load_data: Dict[str, torch.Tensor],
        load_keys: Optional[List[str]] = None,
        detach_all: bool = True,
    ):
        """
        :parameters load_keys: Subset of keys that are loaded from `load_data`.
        """
        if load_keys is None:
            load_keys = load_data.keys()

        self._load_data = {
            k: v.detach() if detach_all else v
            for k, v in load_data.items()
            if k in load_keys
        }
        tensor_sizes = [tensor.size(0) for tensor in self._load_data.values()]
        if len(set(tensor_sizes)) != 1:
            raise ValueError("Tensors to dataset are not of the same shape")
        self._dataset_len = tensor_sizes[0]

    @property
    def all_data(self):
        return self._load_data

    def get_data(self, k: str) -> torch.Tensor:
        return self._load_data[k]

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        return {k: v[idx] for k, v in self._load_data.items()}

    def __len__(self) -> int:
        return self._dataset_len


def extract_next_tensor(dataset: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    obs = dataset["observations"].detach()

    final_final_obs = dataset["infos"][-1]["final_obs"]

    next_obs = torch.cat([obs[1:], final_final_obs.unsqueeze(0)], 0)
    num_eps = 1
    for i in range(obs.shape[0] - 1):
        cur_info = dataset["infos"][i]
        if "final_obs" in cur_info:
            num_eps += 1
            next_obs[i] = cur_info["final_obs"].detach()
    masks = ~(dataset["terminals"].bool())

    num_terminals = masks.size(0) - masks.sum()
    if num_eps != num_terminals.sum():
        raise ValueError(
            f"Inconsistency in # of episodes {num_eps} vs {dataset['terminals'].sum()}"
        )
    dataset["next_obs"] = next_obs.detach()

    return dataset


def extract_final_obs(
    obs: torch.Tensor, masks: torch.Tensor, final_obs: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    :param obs: Shape (N, ...)
    :param masks: Shape (N, ...)
    :param final_obs: Shape (N-1, ...)

    :returns: obs, next_obs, masks all of shape (N-1, ...)
    """
    cur_obs = obs[:-1]
    masks = masks[1:]
    next_obs = (masks * obs[1:]) + ((1 - masks) * final_obs)
    return cur_obs, next_obs, masks



from torch import distributions as pyd


class FixedCategorical(pyd.Categorical):
    def sample(self):
        return super().sample().unsqueeze(-1)

    def log_probs(self, actions):
        return (
            super()
            .log_prob(actions.squeeze(-1))
            .view(actions.size(0), -1)
            .sum(-1)
            .unsqueeze(-1)
        )

    def mode(self):
        return self.probs.argmax(dim=-1, keepdim=True)


class FixedNormal(pyd.Normal):
    def log_probs(self, actions):
        return super().log_prob(actions).sum(-1, keepdim=True)

    def entropy(self):
        return super().entropy().sum(-1)

    def mode(self):
        return self.mean



import numpy as np
from gym import spaces

from .vec_env import VecEnvWrapper


class VecFrameStack(VecEnvWrapper):
    def __init__(self, venv, nstack):
        self.venv = venv
        self.nstack = nstack
        wos = venv.observation_space  # wrapped ob space
        low = np.repeat(wos.low, self.nstack, axis=-1)
        high = np.repeat(wos.high, self.nstack, axis=-1)
        self.stackedobs = np.zeros((venv.num_envs,) + low.shape, low.dtype)
        observation_space = spaces.Box(
            low=low, high=high, dtype=venv.observation_space.dtype
        )
        VecEnvWrapper.__init__(self, venv, observation_space=observation_space)

    def step_wait(self):
        obs, rews, news, infos = self.venv.step_wait()
        self.stackedobs = np.roll(self.stackedobs, shift=-1, axis=-1)
        for (i, new) in enumerate(news):
            if new:
                self.stackedobs[i] = 0
        self.stackedobs[..., -obs.shape[-1] :] = obs
        return self.stackedobs, rews, news, infos

    def reset(self):
        obs = self.venv.reset()
        self.stackedobs[...] = 0
        self.stackedobs[..., -obs.shape[-1] :] = obs
        return self.stackedobs



import os.path as osp

import setuptools

cur_dir = osp.dirname(osp.realpath(__file__))
requirementPath = osp.join(cur_dir, "requirements.txt")
install_requires = []
with open(requirementPath) as f:
    install_requires = f.read().splitlines()

setuptools.setup(
    name="rl-exp-utils",
    version="0.12",
    author="Andrew Szot",
    author_email="andrewszot32@gmail.com",
    description="Library for RL research.",
    url="https://github.com/aszot/rl-utils",
    install_requires=install_requires,
    packages=setuptools.find_packages(),
)



import datetime
import os
import os.path as osp
import string
import time
from collections import defaultdict, deque
from typing import Any, Dict, List, Set, Union

import numpy as np
import torch.nn as nn
from omegaconf import DictConfig

from rl_utils.common.core_utils import compress_and_filter_dict

LoggerCfgType = Union[Dict[str, Any], DictConfig]


class Logger:
    def __init__(
        self,
        seed: int,
        log_dir: str,
        vid_dir: str,
        save_dir: str,
        full_cfg: LoggerCfgType,
        smooth_len: int = 1,
        run_name: str = "",
        **kwargs,
    ):
        """
        :param run_name: If empty string then a run name will be auto generated.
        """
        self._create_run_name(run_name, seed)

        self.log_dir = log_dir
        if self.log_dir != "":
            self.log_dir = osp.join(self.log_dir, self.run_name)
            if not osp.exists(self.log_dir):
                os.makedirs(self.log_dir)

        self.vid_dir = vid_dir
        if self.vid_dir != "":
            self.vid_dir = osp.join(self.vid_dir, self.run_name)
            if not osp.exists(self.vid_dir):
                os.makedirs(self.vid_dir)

        self.save_dir = save_dir
        if self.save_dir != "":
            self.save_dir = osp.join(self.save_dir, self.run_name)
            if not osp.exists(self.save_dir):
                os.makedirs(self.save_dir)

        self._step_log_info = defaultdict(lambda: deque(maxlen=smooth_len))

        self.is_printing = True
        self.prev_steps = 0
        self.start = time.time()
        self._clear_keys: Set[str] = set()

    @property
    def save_path(self):
        return self.save_dir

    @property
    def vid_path(self):
        return self.vid_dir

    def disable_print(self):
        self.is_printing = False

    def collect_env_step_info(self, infos: List[Dict[str, Any]]) -> None:
        for inf in infos:
            if "episode" in inf:
                flat_inf = compress_and_filter_dict(inf)
                # Only log at the end of the episode
                for k, v in flat_inf.items():
                    self._step_log_info[k].append(v)

    def collect_infos(
        self, info: Dict[str, float], prefix: str = "", no_rolling_window: bool = False
    ) -> None:
        for k, v in info.items():
            self.collect_info(k, v, prefix, no_rolling_window)

    def collect_info(
        self, k: str, value: float, prefix: str = "", no_rolling_window: bool = False
    ) -> None:
        """
        :param no_rolling_window: If true, then only the most recent logged
            value will be displayed with a call to `self.interval_log`. This is for
            metrics that should not be averaged.
        """
        use_k = prefix + k
        if no_rolling_window:
            self._step_log_info[use_k].clear()
            self._clear_keys.add(use_k)
        self._step_log_info[use_k].append(value)

    def collect_info_list(self, k: str, values: List[float], prefix: str = "") -> None:
        """
        Collect a list of values for a key.
        """
        for v in values:
            self.collect_info(k, v)

    def _create_run_name(self, run_name, seed):
        if run_name == "":
            d = datetime.datetime.today()
            date_id = "%i%i" % (d.month, d.day)

            chars = list(
                string.ascii_uppercase + string.digits + string.ascii_lowercase
            )
            rnd_id = np.random.RandomState().choice(chars, 6)
            rnd_id = "".join(rnd_id)

            self.run_name = f"{date_id}-{seed}-{rnd_id}"
        else:
            self.run_name = run_name
        print(f"Assigning full prefix {self.run_name}")

    def log_vals(self, key_vals, step_count):
        """
        Log key value pairs to whatever interface.
        """

    def collect_img(self, k: str, img_path: str, prefix: str = ""):
        """
        Log an image
        :param img_path: Full path to the image.
        """

    def watch_model(self, model: nn.Module):
        """
        :param model: the set of parameters to watch
        """

    def interval_log(self, update_count: int, processed_env_steps: int) -> None:
        """
        Printed FPS is all inclusive of updates, evaluations, logging and everything.
        This is NOT the environment FPS.
        :param update_count: The number of updates.
        :param processed_env_steps: The number of environment samples processed.
        """
        end = time.time()

        fps = int((processed_env_steps - self.prev_steps) / (end - self.start))
        self.prev_steps = processed_env_steps
        num_eps = len(self._step_log_info.get("episode.reward", []))

        log_dat = {}
        for k, v in self._step_log_info.items():
            if isinstance(v, deque):
                log_dat[k] = np.mean(v)
            else:
                log_dat[k] = v

        for k in self._clear_keys:
            del self._step_log_info[k]
        self._clear_keys.clear()

        if self.is_printing:
            print("")
            print(f"Updates {update_count}, Steps {processed_env_steps}, FPS {fps}")
            print(f"Over the last {num_eps} episodes:")

            # Print log values from the updater if requested.
            for k, v in log_dat.items():
                print(f"    - {k}: {v}")

        # Log all values
        log_dat["fps"] = fps
        self.log_vals(log_dat, processed_env_steps)
        self.start = end

    def close(self):
        pass



import numpy as np


class RunningMeanStd:
    # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
    def __init__(self, epsilon=1e-4, shape=()):
        # Note: This was float64 earlier, but we changed it to float32
        self.mean = np.zeros(shape, "float32")
        self.var = np.ones(shape, "float32")
        self.count = epsilon

    def update(self, x):
        batch_mean = np.mean(x, axis=0, dtype="float32")
        batch_var = np.var(x, axis=0, dtype="float32")
        batch_count = x.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(self, batch_mean, batch_var, batch_count):
        self.mean, self.var, self.count = update_mean_var_count_from_moments(
            self.mean, self.var, self.count, batch_mean, batch_var, batch_count
        )

    def __str__(self):
        return "Mean: %s, Var: %s, Count: %i" % (
            str(list(self.mean)),
            str(list(self.var)),
            self.count,
        )


def update_mean_var_count_from_moments(
    mean, var, count, batch_mean, batch_var, batch_count
):
    delta = batch_mean - mean
    tot_count = count + batch_count

    new_mean = mean + delta * batch_count / tot_count
    m_a = var * count
    m_b = batch_var * batch_count
    M2 = m_a + m_b + np.square(delta, dtype="float32") * count * batch_count / tot_count
    new_var = M2 / tot_count
    new_count = tot_count

    return new_mean, new_var, new_count



from typing import Dict, Optional, Type


class EnvRegistry:
    def __init__(self):
        self._envs: Dict[str, Type] = {}

    def register_env(self, register_name: str):
        def wrap(to_register_cls: Type):
            self._envs[register_name] = to_register_cls
            return to_register_cls

        return wrap

    def search_env(self, name: str) -> Optional[Type]:
        return self._envs.get(name, None)


full_env_registry = EnvRegistry()



from omegaconf import DictConfig, OmegaConf

from rl_utils.logging.base_logger import Logger, LoggerCfgType

try:
    import wandb
except ImportError:
    wandb = None


class WbLogger(Logger):
    """
    Logger for logging to the weights and W&B online service.
    """

    def __init__(
        self,
        wb_proj_name: str,
        wb_entity: str,
        seed: int,
        log_dir: str,
        vid_dir: str,
        save_dir: str,
        full_cfg: LoggerCfgType,
        smooth_len: int = 1,
        run_name: str = "",
        group_name: str = "",
        **kwargs,
    ):
        """
        :parameter run_name: If empty string then a random run name is assigned.
        :parameter group_name: If empty string then no group name is used.
        """
        if wandb is None:
            raise ImportError("Wandb is not installed")

        super().__init__(
            seed, log_dir, vid_dir, save_dir, full_cfg, smooth_len, run_name
        )
        if wb_proj_name == "" or wb_entity == "":
            raise ValueError(
                f"Must specify W&B project and entity name {wb_proj_name}, {wb_entity}"
            )

        self.wb_proj_name = wb_proj_name
        self.wb_entity = wb_entity
        self.wandb = self._create_wandb(full_cfg, group_name)

    def log_vals(self, key_vals, step_count):
        wandb.log(key_vals, step=int(step_count))

    def watch_model(self, model):
        wandb.watch(model)

    def _create_wandb(self, full_cfg: LoggerCfgType, group_name: str):
        if group_name == "":
            group_name = None
        if isinstance(full_cfg, DictConfig):
            full_cfg = OmegaConf.to_container(full_cfg, resolve=True)

        self.run = wandb.init(
            project=self.wb_proj_name,
            name=self.run_name,
            entity=self.wb_entity,
            group=group_name,
            config=full_cfg,
        )
        return wandb

    def collect_img(self, k: str, img_path: str, prefix: str = ""):
        use_k = prefix + k
        self._step_log_info[use_k] = wandb.Image(img_path)
        self._clear_keys.add(use_k)

    def close(self):
        self.run.finish()



import torch

from rl_utils.models import SimpleCNN, build_rnn_state_encoder


def test_simple_cnn():
    model = SimpleCNN([64, 64, 3], 128)
    model(torch.randn(128, 3, 64, 64))

    obs_shape = {"rgb": [64, 64, 3], "other_state": [9]}
    model = SimpleCNN(obs_shape, 128)
    model({k: torch.randn(128, v[-1], *v[:2]) for k, v in obs_shape.items()})


def test_rnn():
    build_rnn_state_encoder(256, 128, "GRU", 2)

    build_rnn_state_encoder(256, 128, "LSTM", 2)



import os
import os.path as osp
from collections import defaultdict
from typing import Dict, Optional

import numpy as np
import torch
import torch.nn as nn

from rl_utils.common.core_utils import compress_and_filter_dict
from rl_utils.common.viz_utils import save_mp4
from rl_utils.envs.vec_env.vec_env import VecEnv
from rl_utils.interfaces import BasePolicy


class Evaluator:
    """
    Dataset save format is meant to be consistent with https://github.com/rail-berkeley/d4rl/blob/master/d4rl/offline_env.py
    """

    def __init__(
        self,
        envs: VecEnv,
        rnn_hxs_dim: int,
        num_render: Optional[int] = None,
        vid_dir: str = "data/vids",
        fps: int = 10,
        save_traj_name: Optional[str] = None,
        **kwargs,
    ):
        """
        :param save_traj_name: The full file path (for example
            "data/trajs/data.pth") to save the evaluated trajectories to.
        :param num_render: If None then every episode will be rendered.
        :param rnn_hxs_dim: The recurrent hidden state dimension.
        """
        self._envs = envs
        self._rnn_hxs_dim = rnn_hxs_dim
        self._num_render = num_render
        self._vid_dir = vid_dir
        self._fps = fps
        self._should_save_trajs = save_traj_name is not None
        self._save_traj_name = save_traj_name

    def _clear_save_trajs(self):
        self._save_trajs_obs = defaultdict(list)
        self._save_trajs_actions = defaultdict(list)
        self._save_trajs_rewards = defaultdict(list)
        self._save_trajs_done = defaultdict(list)
        self._save_trajs_info = defaultdict(list)

        self._all_traj_obs = []
        self._all_traj_actions = []
        self._all_traj_rewards = []
        self._all_traj_done = []
        self._all_traj_info = []

    def _add_transition_to_save(self, env_i, obs, action, reward, done, info):
        self._save_trajs_obs[env_i].append(obs[env_i])
        self._save_trajs_actions[env_i].append(action[env_i])
        self._save_trajs_rewards[env_i].append(reward[env_i])
        self._save_trajs_done[env_i].append(done[env_i])
        self._save_trajs_info[env_i].append(info[env_i])

    def _flush_trajectory_to_save(self, env_i):
        self._all_traj_obs.extend(self._save_trajs_obs[env_i])
        self._all_traj_actions.extend(self._save_trajs_actions[env_i])
        self._all_traj_rewards.extend(self._save_trajs_rewards[env_i])
        self._all_traj_done.extend(self._save_trajs_done[env_i])
        self._all_traj_info.extend(self._save_trajs_info[env_i])

        self._save_trajs_obs[env_i].clear()
        self._save_trajs_actions[env_i].clear()
        self._save_trajs_rewards[env_i].clear()
        self._save_trajs_done[env_i].clear()
        self._save_trajs_info[env_i].clear()

    @property
    def eval_trajs_obs(self):
        return self._all_traj_obs

    @property
    def eval_trajs_dones(self):
        return self._all_traj_done

    def _save_trajs(self):
        assert self._save_traj_name is not None
        obs = torch.stack(self._all_traj_obs, dim=0).detach()
        actions = torch.stack(self._all_traj_actions, dim=0).detach()
        rewards = torch.stack(self._all_traj_rewards, dim=0).detach()
        terminals = torch.stack(self._all_traj_done, dim=0).detach()

        num_steps = obs.shape[0]
        assert (
            actions.shape[0] == num_steps and len(actions.shape) == 2
        ), f"Action shape wrong {actions.shape}"

        rewards = rewards.view(-1)
        terminals = terminals.view(-1)
        assert rewards.size(0) == num_steps, f"Reward is wrong shape {rewards.shape}"
        assert (
            terminals.size(0) == num_steps
        ), f"Terminals is wrong shape {terminals.shape}"

        os.makedirs(osp.dirname(self._save_traj_name), exist_ok=True)

        torch.save(
            {
                "observations": obs,
                "actions": actions,
                "rewards": rewards,
                "terminals": terminals.float(),
                "infos": self._all_traj_info,
            },
            self._save_traj_name,
        )
        print(f"Saved trajectories to {self._save_traj_name}")
        self._clear_save_trajs()

    def evaluate(
        self, policy: BasePolicy, num_episodes: int, eval_i: int
    ) -> Dict[str, float]:
        self._clear_save_trajs()

        if isinstance(policy, nn.Module):
            device = next(policy.parameters()).device
        else:
            device = torch.device("cpu")

        num_envs = self._envs.num_envs
        obs = self._envs.reset()
        rnn_hxs = torch.zeros(num_envs, self._rnn_hxs_dim).to(device)
        eval_masks = torch.zeros(num_envs, 1, device=device)

        evals_per_proc = num_episodes // num_envs
        left_over_evals = num_episodes % num_envs
        num_evals = [evals_per_proc for _ in range(num_envs)]
        num_evals[-1] += left_over_evals

        all_frames = []
        accum_stats = defaultdict(list)
        total_evaluated = 0
        if self._num_render is None:
            num_render = num_episodes
        else:
            num_render = self._num_render
        with torch.no_grad():
            while sum(num_evals) != 0:
                act_data = policy.act(obs, rnn_hxs, eval_masks, is_eval=True)
                next_obs, rewards, done, info = self._envs.step(act_data["actions"])
                rnn_hxs = act_data["hxs"]

                if total_evaluated < num_render:
                    frames = self._envs.render(mode="rgb_array")
                    all_frames.append(frames)

                for env_i in range(num_envs):
                    self._add_transition_to_save(
                        env_i, obs, act_data["actions"], rewards, done, info
                    )

                    if done[env_i]:
                        total_evaluated += 1
                        if num_evals[env_i] > 0:
                            self._flush_trajectory_to_save(env_i)
                            for k, v in compress_and_filter_dict(info[env_i]).items():
                                accum_stats[k].append(v)
                            num_evals[env_i] -= 1
                obs = next_obs

        if len(all_frames) > 0:
            save_mp4(all_frames, self._vid_dir, f"eval_{eval_i}", self._fps)

        if self._should_save_trajs:
            self._save_trajs()

        return {k: np.mean(v) for k, v in accum_stats.items()}



import abc
from typing import Any, Dict, Union

import gym.spaces as spaces
import numpy as np
import torch


class BasePolicy(abc.ABC):
    """
    Defines core policy functionality for policies to be compatible with multiple code bases.
    """

    def act(
        self,
        obs: Union[torch.Tensor, Dict[str, torch.Tensor]],
        hxs: torch.Tensor,
        masks: torch.Tensor,
        is_eval: bool = False,
    ) -> Dict[str, Any]:
        """
        :returns: Dictionary at least with keys (can include more information
            as needed, such as action log probabilities):
            "actions": torch.Tensor of shape [batch_size, action_dim]
            "hxs": torch.Tensor of shape [batch_size, D] for hidden dimension size D.
        """
        raise NotImplementedError()


class RandomPolicy(BasePolicy):
    def __init__(self, action_space: spaces.Space):
        self._action_space = action_space

    def act(self, obs, hxs, masks, is_eval=False) -> Dict[str, Any]:
        action = torch.tensor(
            np.array([self._action_space.sample() for _ in range(obs.size(0))])
        )

        return {"actions": action, "hxs": hxs}
